<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace SLAModule\Command;


use App\Lib\Traits\DatabasePartitionsTrait;
use Cake\Console\Arguments;
use Cake\Console\Command;
use Cake\Console\ConsoleIo;
use Cake\Console\ConsoleOptionParser;
use Cake\Log\Log;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\Interfaces\CronjobInterface;
use SLAModule\Lib\Sla;
use SLAModule\Lib\SlaHost;
use SLAModule\Lib\SlaService;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsLogTable;
use SLAModule\Model\Table\SlaAvailabilityStatusServicesLogTable;
use SLAModule\Model\Table\SlaHostOutagesTable;
use SLAModule\Model\Table\SlaServiceOutagesTable;
use SLAModule\Model\Table\SlaSettingsTable;
use SLAModule\Model\Table\SlasTable;

/**
 * Class SlaCalculateAvailabilityCommand
 * @package SLAModule\Command
 */
class SlaCalculateAvailabilityCommand extends Command implements CronjobInterface {

    use DatabasePartitionsTrait;

    /**
     * @var ConsoleIo
     */
    private $io;

    /**
     * Hook method for defining this command's option parser.
     *
     * @see https://book.cakephp.org/3.0/en/console-and-shells/commands.html#defining-arguments-and-options
     *
     * @param \Cake\Console\ConsoleOptionParser $parser The parser to be defined
     * @return \Cake\Console\ConsoleOptionParser The built parser.
     */
    public function buildOptionParser(ConsoleOptionParser $parser): ConsoleOptionParser {
        $parser = parent::buildOptionParser($parser);

        return $parser;
    }

    /**
     * @param Arguments $args
     * @param ConsoleIo $io
     * @return int|void|null
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     * @throws \GuzzleHttp\Exception\GuzzleException
     * @throws \Exception
     */
    public function execute(Arguments $args, ConsoleIo $io) {
        $this->io = $io;

        $this->checkAndCreatPartitions($io);

        $io->out('Calculate SLA reports ...');

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        $slas = $SlasTable->getSlasForCronjob();
        foreach ($slas as $sla) {
            /** @var \SLAModule\Model\Entity\Sla $sla */
            $Sla = new Sla($sla);
            $result = $Sla->canSlaBeCalculated();
            if ($result === true) {
                $io->out(sprintf('Calculating: %s', $sla->name), 0);
                $hosts = $SlasTable->getHostsBySlaId($sla->get('id'));
                foreach ($hosts as $host) {
                    $SlaHost = new SlaHost($sla, $host);
                    $SlaHost->calculateAvailabilityForHost();

                    foreach ($host->get('services') as $service) {
                        $SlaService = new SlaService($sla, $service);
                        $SlaService->calculateAvailabilityForService();
                    }

                    $SlaHost->updateSlaServiceStatus();
                }
                $io->success('   Ok');
            }
        }

        $io->hr();
    }

    public function checkAndCreatPartitions(ConsoleIo $io) {
        /***********************************************/
        /************ UNIX TIMESTAMP TABLES ************/
        /***********************************************/

        /** @var SlaHostOutagesTable $SlaHostOutagesTable */
        $SlaHostOutagesTable = TableRegistry::getTableLocator()->get('SLAModule.SlaHostOutages');
        /** @var SlaServiceOutagesTable $SlaServiceOutagesTable */
        $SlaServiceOutagesTable = TableRegistry::getTableLocator()->get('SLAModule.SlaServiceOutages');


        // Make sure that the initial partition exists
        // This is for tables using a unix timestamp as partition by value
        foreach ([$SlaHostOutagesTable, $SlaServiceOutagesTable] as $Table) {
            /** @var Table $Table */
            $Connection = $Table->getConnection();

            $partitions = $this->getPartitionsByTable($Table);
            if (empty($partitions)) {
                // This table has no partitions yet (fresh installation of the module)
                // Execute "ALTER TABLE" to create the initial partition "p_max"
                // This way we can still use cakephp/migrations to manage the schema
                $this->alterTableAndCreateFirstPartitionByUnixtimestampUnsafe($Table, 'start_time');
                $partitions = $this->getPartitionsByTable($Table);
            }

            //Check if partition for current week exists
            $currentMysqlPartitionStartDate = strtotime('00:00:00 next monday');
            $currentMysqlPartitionStartDate = intdiv($currentMysqlPartitionStartDate, 86400);

            $currentPartitionName = 'p_' . date('o_W');

            $io->out('Checking for partition ' . $currentPartitionName . ' in Table "' . $Table->getTable() . '"...', 0);
            if (!in_array($currentPartitionName, $partitions, true)) {
                try {                                                                                                                                                                                                               // < next monday 00:00
                    $Connection->execute("ALTER TABLE " . $Connection->config()['database'] . "." . $Table->getTable() . " REORGANIZE PARTITION p_max INTO (PARTITION " . $currentPartitionName . " VALUES LESS THAN (" . $currentMysqlPartitionStartDate . "), PARTITION p_max values LESS THAN (MAXVALUE));");
                    $io->success('   Ok');
                } catch (\Exception $e) {
                    Log::error('SlaCalculateAvailabilityCommand: MySQL Error: ' . $e->getMessage());
                }
            } else {
                $io->success('   Ok');
            }


            // Attention: look close before you change the date parameters!!!
            // Think that there's coming a new year on 31.12 !!!
            // TO_DAYS('2014-12-29 00:00:00')
            $mysqlPartitionStartDate = strtotime('00:00:00 next monday +1 week');
            $mysqlPartitionStartDate = intdiv($mysqlPartitionStartDate, 86400);
            //p_2015_01
            $newPartitionName = 'p_' . date('o_W', strtotime('next monday'));

            $io->out('Checking for partition ' . $newPartitionName . ' in Table "' . $Table->getTable() . '"...', 0);

            //Checking if this partition already exists
            if (!in_array($newPartitionName, $partitions, true)) {
                //This partition does not exist, and we need to create it
                $io->out('');
                $io->info('Partition does not exists, will create it....', 0);

                $Connection->execute("ALTER TABLE " . $Connection->config()['database'] . "." . $Table->getTable() . " REORGANIZE PARTITION p_max INTO (PARTITION " . $newPartitionName . " VALUES LESS THAN (" . $mysqlPartitionStartDate . "), PARTITION p_max values LESS THAN (MAXVALUE));");
                $io->success('   Ok');
                //ALTER TABLE ptn_test DROP PARTITION p_2015_01
            } else {
                $io->success('   Ok');
            }
            $this->cleanupPartition($Table, $partitions, $io);
            $io->hr();
        }

        /******************************************/
        /************ DATE TIME TABLES ************/
        /******************************************/

        /** @var  SlaAvailabilityStatusHostsLogTable $SlaAvailabilityStatusHostsLogTable */
        $SlaAvailabilityStatusHostsLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHostsLog');
        /** @var  SlaAvailabilityStatusServicesLogTable $SlaAvailabilityStatusServicesLogTable */
        $SlaAvailabilityStatusServicesLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServicesLog');

        // Make sure that the initial partition exists
        // This is for tables using MySLQ DATETIME as partition by value
        foreach ([$SlaAvailabilityStatusHostsLogTable, $SlaAvailabilityStatusServicesLogTable] as $Table) {
            /** @var Table $Table */
            $Connection = $Table->getConnection();

            $partitions = $this->getPartitionsByTable($Table);
            if (empty($partitions)) {
                // This table has no partitions yet (fresh installation of the module)
                // Execute "ALTER TABLE" to create the initial partition "p_max"
                // This way we can still use cakephp/migrations to manage the schema
                $this->alterTableAndCreateFirstPartitionByDatetimeUnsafe($Table, 'evaluation_start');
                $partitions = $this->getPartitionsByTable($Table);
            }

            //Check if partition for current week exists
            $currentMysqlPartitionStartDate = date('Y-m-d H:i:s', strtotime('00:00:00 next monday'));
            $currentPartitionName = 'p_' . date('o_W');

            $io->out('Checking for partition ' . $currentPartitionName . ' in Table "' . $Table->getTable() . '"...', 0);
            if (!in_array($currentPartitionName, $partitions, true)) {
                try {
                    $Connection->execute("ALTER TABLE " . $Connection->config()['database'] . "." . $Table->getTable() . " REORGANIZE PARTITION p_max INTO (PARTITION " . $currentPartitionName . " VALUES LESS THAN (TO_DAYS('" . $currentMysqlPartitionStartDate . "')), PARTITION p_max values LESS THAN (MAXVALUE));");
                    $io->success('   Ok');
                } catch (\Exception $e) {
                    Log::error('SlaCalculateAvailabilityCommand: MySQL Error: ' . $e->getMessage());
                }
            } else {
                $io->success('   Ok');
            }


            // Attention: look close before you change the date parameters!!!
            // Think that there's coming a new year on 31.12 !!!
            // TO_DAYS('2014-12-29 00:00:00')
            $mysqlPartitionStartDate = date('Y-m-d H:i:s', strtotime('00:00:00 next monday +1 week'));
            //p_2015_01
            $newPartitionName = 'p_' . date('o_W', strtotime('next monday'));

            $io->out('Checking for partition ' . $newPartitionName . ' in Table "' . $Table->getTable() . '"...', 0);

            //Checking if this partition already exists
            if (!in_array($newPartitionName, $partitions, true)) {
                //This partition does not exist, and we need to create it
                $io->out('');
                $io->info('Partition does not exists, will create it....', 0);

                $Connection->execute("ALTER TABLE " . $Connection->config()['database'] . "." . $Table->getTable() . " REORGANIZE PARTITION p_max INTO (PARTITION " . $newPartitionName . " VALUES LESS THAN (TO_DAYS('" . $mysqlPartitionStartDate . "')), PARTITION p_max values LESS THAN (MAXVALUE));");
                $io->success('   Ok');
                //ALTER TABLE ptn_test DROP PARTITION p_2015_01
            } else {
                $io->success('   Ok');
            }
            $this->cleanupPartition($Table, $partitions, $io);
            $io->hr();
        }
    }

    public function cleanupPartition(Table $Table, array $partitions, ConsoleIo $io) {
        $io->out('Checking age of partitions...');

        /** @var SlaSettingsTable $SlaSettingsTable */
        $SlaSettingsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaSettings');
        $sla_settings = $SlaSettingsTable->getSettings();

        $maxAgeInWeeks = $sla_settings['max_age'] ?? 53;

        $maxAgeTimestamp = strtotime('00:00:00 last monday -' . ($maxAgeInWeeks + 1) . ' week');

        //Valide date
        if ($maxAgeTimestamp) {
            //$partitions = ['p_2014_48', 'p_2014_49', 'p_2014_50', 'p_2014_51', 'p_2015_01', 'p_2015_02', 'p_2015_03'];
            $comparePartitionName = 'p_' . date('o_W', $maxAgeTimestamp);


            if (is_array($partitions)) {
                foreach ($partitions as $partition) {
                    //NEVER EVER DROP THIS PARTITION!!!
                    if ($partition == 'p_max') {
                        continue;
                    }
                    if (!version_compare($comparePartitionName, $partition, '<')) {
                        $io->setStyle('red', ['text' => 'red', 'blink' => false]);

                        $io->out('<red>Drop partition ' . $partition . '</red>');
                        $this->dropPartitionByNameUnsafe($Table, $partition);
                    }
                }
            }

            $io->success('Cleanup done');
        }
    }

}
