<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace SLAModule\Controller;

use App\Model\Table\ContainersTable;
use App\Model\Table\HostsTable;
use Cake\Datasource\ResultSetInterface;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use itnovum\openITCOCKPIT\Core\HostConditions;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use SLAModule\Model\Entity\HostMappingRule;
use SLAModule\Model\Table\HostMappingRulesTable;
use SLAModule\Model\Table\SlasTable;

/**
 * HostMappingRules Controller
 *
 * @property HostMappingRulesTable $HostMappingRules
 * @method HostMappingRule[]|ResultSetInterface paginate($object = null, array $settings = [])
 */
class HostMappingRulesController extends AppController {

    public function assignToHosts($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }
        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        if (!$SlasTable->existsById($id)) {
            throw new NotFoundException(__('SLA not found'));
        }

        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        /** @var HostMappingRulesTable $HostMappingRulesTable */
        $HostMappingRulesTable = TableRegistry::getTableLocator()->get('SLAModule.HostMappingRules');

        $sla = $SlasTable->get($id, [
            'contain' => [
                'HostMappingRules'
            ]
        ]);
        if (!$this->allowedByContainerId($sla['container_id'])) {
            $this->render403();
            return;
        }
        if ($this->request->is('get') && $this->isAngularJsRequest()) {

            $sla['allowEdit'] = $this->isWritableContainer($sla['container_id']);

            $this->set('sla', $sla);
            $this->viewBuilder()->setOption('serialize', ['sla']);
            return;
        }

        if ($this->request->is('post') || $this->request->is('put')) {
            $requestData = $this->request->getData(null, []);
            $containerId = $sla->get('container_id');

            $GenericFilter = new GenericFilter($this->request);
            $GenericFilter->setFilters([
                'rlike'    => [
                    'Hosts.name',
                    'Hosts.keywords'
                ],
                'notrlike' => [
                    'Hosts.not_keywords'
                ],
            ]);

            if ($containerId == ROOT_CONTAINER) {
                //Don't panic! Only root users can edit /root objects ;)
                //So no loss of selected hosts/host templates
                $containerIds = $ContainersTable->resolveChildrenOfContainerIds(ROOT_CONTAINER, true);
            } else {
                $containerIds = $ContainersTable->resolveChildrenOfContainerIds($containerId, true);
                $containerIds = array_merge($containerIds, [ROOT_CONTAINER, $containerId]);
            }

            $hostMappingRule = $HostMappingRulesTable->getHostMappingRuleBySlaId($id);

            if (empty($hostMappingRule)) {
                $hostMappingRule = $HostMappingRulesTable->newEntity($requestData);
            } else {
                $hostMappingRule = $HostMappingRulesTable->patchEntity($hostMappingRule, $requestData);
            }

            if ($requestData['save_options'] === false) {
                if ($hostMappingRule->hasErrors()) {
                    $this->response = $this->response->withStatus(400);
                    $this->set('error', $hostMappingRule->getErrors());
                    $this->viewBuilder()->setOption('serialize', ['error']);
                    return;
                }
            }

            $HostMappingRulesTable->save($hostMappingRule);

            if ($hostMappingRule->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $hostMappingRule->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            $HostCondition = new HostConditions();
            $HostCondition->setContainerIds($containerIds);

            //list host_id => hosttemplate_sla_id
            $hostIds = $HostMappingRulesTable->getHostsIndex(
                $GenericFilter,
                $HostCondition,
                $hostMappingRule->get('only_unassigned'),
                'list'
            );

            $hostIdsForCleanUp = [];
            if (!empty($hostIds)) {
                //update affected hosts
                /** @var HostsTable HostsTable */
                $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
                foreach ($hostIds as $hostId => $slaIdFromHosttemplate) {
                    $host = $HostsTable->get($hostId);
                    $oldSlaId = $host->get('sla_id');
                    if (!is_null($oldSlaId) && $oldSlaId != $id) {
                        $hostIdsForCleanUp[] = $hostId;
                    } else if (is_null($oldSlaId) && !is_null($slaIdFromHosttemplate) && $slaIdFromHosttemplate != $id) {
                        $hostIdsForCleanUp[] = $hostId;
                    }
                    $host->set('sla_id', ($id == $slaIdFromHosttemplate) ? null : $id);
                    $HostsTable->save($host);

                }
                if (!empty($hostIdsForCleanUp)) {
                    $SlasTable->deleteSlaRecordsByHostIds($hostIdsForCleanUp);
                }
            }

            $this->set('sla', $hostMappingRule);
            $this->set('hostMappingRule', $sla);
            $this->viewBuilder()->setOption('serialize', ['sla', 'hostMappingRule']);
        }
    }

    public function loadHostsBySlaId() {
        if (!$this->isAngularJsRequest()) {
            throw new MethodNotAllowedException();
        }
        $slaId = $this->request->getQuery('slaId');
        $onlyUnassigned = $this->request->getQuery('onlyUnassigned', false) === 'true';

        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        /** @var HostMappingRulesTable $HostMappingRulesTable */
        $HostMappingRulesTable = TableRegistry::getTableLocator()->get('SLAModule.HostMappingRules');

        $sla = $SlasTable->get($slaId);
        $containerId = $sla->get('container_id');

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like'     => [
                'hostdescription'
            ],
            'rlike'    => [
                'Hosts.name',
                'Hosts.keywords'
            ],
            'notrlike' => [
                'Hosts.not_keywords'
            ],
        ]);


        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());

        if ($containerId == ROOT_CONTAINER) {
            //Don't panic! Only root users can edit /root objects ;)
            //So no loss of selected hosts/host templates
            $containerIds = $ContainersTable->resolveChildrenOfContainerIds(ROOT_CONTAINER, true);
        } else {
            $containerIds = $ContainersTable->resolveChildrenOfContainerIds($containerId, true);
            $containerIds = array_merge($containerIds, [ROOT_CONTAINER, $containerId]);
        }
        $HostCondition = new HostConditions();
        $HostCondition->setContainerIds($containerIds);

        $hosts = $HostMappingRulesTable->getHostsIndex(
            $GenericFilter,
            $HostCondition,
            $onlyUnassigned,
            'all',
            $PaginateOMat
        );
        foreach ($hosts as $key => $host) {
            $activeSlaId = is_null($host['sla_id']) ? $host['hosttemplate']['sla_id'] : $host['sla_id'];
            $hosts[$key]['sla'] = null;
            if ($activeSlaId) {
                $slaFromHost = $SlasTable->get($activeSlaId);
                $hosts[$key]['sla'] = $slaFromHost;
                $hosts[$key]['sla']['allowEdit'] = $this->isWritableContainer($slaFromHost['container_id']);
            }
        }

        $this->set('hosts', $hosts);
        $this->viewBuilder()->setOption('serialize', ['hosts']);
    }
}
