<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace SLAModule\Controller;

use App\Model\Table\ContainersTable;
use App\Model\Table\HostsTable;
use App\Model\Table\HosttemplatesTable;
use App\Model\Table\ServicesTable;
use App\Model\Table\WidgetsTable;
use Cake\Datasource\ResultSetInterface;
use Cake\Http\Exception\MethodNotAllowedException;
use Cake\Http\Exception\NotFoundException;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use itnovum\openITCOCKPIT\Core\AngularJS\Api;
use itnovum\openITCOCKPIT\Core\CalendarTime;
use itnovum\openITCOCKPIT\Core\ValueObjects\User;
use itnovum\openITCOCKPIT\Core\Views\ContainerPermissions;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use SLAModule\Form\SlaReportForm;
use SLAModule\Lib\CsvReport;
use SLAModule\Lib\SlaReport;
use SLAModule\Model\Entity\Sla;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsLogTable;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsTable;
use SLAModule\Model\Table\SlaAvailabilityStatusServicesLogTable;
use SLAModule\Model\Table\SlaAvailabilityStatusServicesTable;
use SLAModule\Model\Table\SlasTable;
use SLAModule\Reports\SlaDaterangesCreator;

/**
 * Slas Controller
 *
 * @method Sla[]|ResultSetInterface paginate($object = null, array $settings = [])
 */
class SlasController extends AppController {
    public function index() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }

        /** @var User $User */
        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        /** @var ContainersTable $ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like' => [
                'Slas.name',
                'Slas.description'
            ]
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        /** @var SlaAvailabilityStatusHostsTable $SlaAvailabilityStatusHostsTable */
        $SlaAvailabilityStatusHostsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHosts');

        $slas = $SlasTable->getSlasIndex($GenericFilter, $PaginateOMat, $MY_RIGHTS);
        foreach ($slas as $index => $sla) {
            $slas[$index]['container'] = '/' . $ContainersTable->treePath($sla['container_id']);
            $slas[$index]['allowEdit'] = $this->isWritableContainer($sla['container_id']);
            $slas[$index]['start_date'] = !is_null($sla['start_date']) ? $UserTime->format($sla['start_date']) : null;
            $slaStatus = $SlaAvailabilityStatusHostsTable->getCurrentDeterminedAvailabilityBySlaId($sla['id']);
            $slas[$index]['status_percent'] = $slaStatus['determined_availability'];
            $slas[$index]['fulfill'] = ($slaStatus['determined_availability'] >= $sla['minimal_availability']);
        }

        $this->set('slas', $slas);
        $this->viewBuilder()->setOption('serialize', ['slas']);

    }

    public function add() {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        if ($this->request->is('post') && $this->isAngularJsRequest()) {
            $sla = $SlasTable->newEmptyEntity();
            $requestData = $this->request->getData();
            if (!empty($requestData['start_date'])) {
                $requestData['start_date'] = $UserTime->customFormat('Y-m-d', $requestData['start_date']);
            }

            $sla = $SlasTable->patchEntity($sla, $requestData);
            $SlasTable->save($sla);
            if ($sla->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->serializeCake4ErrorMessage($sla);
                return;
            } else {
                //No errors
                $this->serializeCake4Id($sla);
            }
            $this->set('sla', $sla);
            $this->viewBuilder()->setOption('serialize', ['sla']);
        }
    }

    /**
     * @param int|null $id
     */
    public function edit($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        if (!$SlasTable->existsById($id)) {
            throw new NotFoundException(__('SLA not found'));
        }

        $sla = $SlasTable->get($id);
        if (!$this->allowedByContainerId($sla['container_id'])) {
            $this->render403();
            return;
        }

        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            if (!empty($sla['start_date'])) {
                $sla['start_date'] = $UserTime->customFormat('Y-m-d', $sla['start_date']);
            }
            $sla['allowEdit'] = $this->isWritableContainer($sla['container_id']);

            $this->set('sla', $sla);
            $this->viewBuilder()->setOption('serialize', ['sla']);
            return;
        }

        if ($this->request->is('post') || $this->request->is('put')) {
            $requestData = $this->request->getData(null, []);
            if (!empty($requestData['start_date'])) {
                $requestData['start_date'] = $UserTime->customFormat('Y-m-d', $requestData['start_date']);
            }

            $sla = $SlasTable->patchEntity($sla, $requestData);

            $SlasTable->save($sla);

            if ($sla->hasErrors()) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $sla->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }

            $this->set('sla', $sla);
            $this->viewBuilder()->setOption('serialize', ['sla']);
        }
    }


    /**
     * @param null $id
     */
    public function delete($id = null) {
        if (!$this->request->is('post')) {
            throw new MethodNotAllowedException();
        }

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        $id = (int)$id;
        if (!$SlasTable->existsById($id)) {
            throw new NotFoundException(__('SLA not found'));
        }

        $sla = $SlasTable->get($id);
        if (!$this->allowedByContainerId($sla->get('container_id'))) {
            $this->render403();
            return;
        }

        if ($SlasTable->delete($sla)) {
            $this->set('success', true);

            // Remove sla_id from host templates and hosts
            /** @var HosttemplatesTable $HosttemplatesTable */
            $HosttemplatesTable = TableRegistry::getTableLocator()->get('Hosttemplates');
            /** @var HostsTable $HostsTable */
            $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

            $HosttemplatesTable->updateAll(['sla_id' => null], ['sla_id' => $id]);
            $HostsTable->updateAll(['sla_id' => null], ['sla_id' => $id]);


            $this->viewBuilder()->setOption('serialize', ['success']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
    }

    public function timeperiodDetailsTooltip() {
        //Return HTML Template for PaginatorDirective
        return;
    }


    /**
     * @param int|null $id
     */
    public function hosts($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        if (!$SlasTable->existsById($id)) {
            throw new NotFoundException(__('SLA not found'));
        }

        $sla = $SlasTable->get($id);
        if (!$this->allowedByContainerId($sla['container_id'])) {
            $this->render403();
            return;
        }
        $sla['allowEdit'] = $this->isWritableContainer($sla['container_id']);

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like' => [
                'Hosts.name'
            ]
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());


        if ($this->request->is('get') && $this->isAngularJsRequest()) {
            $hosts = $SlasTable->getHostsForOverviewBySlaId(
                $id,
                $HostsTable,
                $GenericFilter,
                $PaginateOMat,
                $MY_RIGHTS
            );

            foreach ($hosts as $key => $host) {
                $allowEdit = true;
                $allowEditHosttemplate = true;
                if ($this->hasRootPrivileges === false) {
                    if (!isset($hostContainers[$host['id']])) {
                        $hostContainers[$host['id']] = $HostsTable->getHostContainerIdsByHostId($host['id']);
                    }
                    $ContainerPermissions = new ContainerPermissions($this->MY_RIGHTS_LEVEL, $hostContainers[$host['id']]);
                    $allowEdit = $ContainerPermissions->hasPermission();
                    $allowEditHosttemplate = $this->isWritableContainer($host['hosttemplate']['container_id']);
                }
                $hosts[$key]['allowEdit'] = $allowEdit;
                $hosts[$key]['hosttemplate']['allowEdit'] = $allowEditHosttemplate;
            }

            $sla['allowEdit'] = $this->isWritableContainer($sla['container_id']);

            $this->set('sla', $sla);
            $this->set('hosts', $hosts);
            $this->set('success', true);
            $this->viewBuilder()->setOption('serialize', ['sla', 'hosts', 'success']);
            return;
        }

        $this->response = $this->response->withStatus(400);
        $this->set('success', false);
        $this->viewBuilder()->setOption('serialize', ['success']);
    }

    /**
     * @param $id
     * @return void
     * @throws \Exception
     */
    public function slaHostInformation($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }
        if (!$this->isAngularJsRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        if (!$HostsTable->existsById($id)) {
            throw new NotFoundException(__('Invalid host'));
        }

        $host = $HostsTable->getHostByIdWithHosttemplate($id);

        $containerIdsToCheck = Hash::extract($host, 'hosts_to_containers_sharing.{n}.id');
        $containerIdsToCheck[] = $host['container_id'];
        if (!$this->allowedByContainerId($containerIdsToCheck, false)) {
            $this->render403();
            return;
        }

        $slaId = !empty($host['sla_id']) ? $host['sla_id'] : $host['hosttemplate']['sla_id'];

        $User = new User($this->getUser());
        $UserTime = UserTime::fromUser($User);

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        $sla = $SlasTable->get($slaId, [
            'contain' => [
                'Timeperiods' => 'TimeperiodTimeranges'
            ]
        ]);

        if ($sla === null) {
            $this->set('response', []);
            $this->viewBuilder()->setOption('serialize', ['response']);
            return;
        }

        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        //Load containers information
        if (array_key_exists($sla['container_id'], $this->MY_RIGHTS_LEVEL)) {
            $mainContainer = $ContainersTable->getTreePathForBrowser($sla['container_id'], $this->MY_RIGHTS_LEVEL);
        } else {
            //The user only see this host via host sharing
            //We need to "fake" a primary container because the user has no permissions to the real
            //primary container
            $mainContainer = $ContainersTable->getFakePrimaryContainerForHostBrowserDisplay(
                $sla['container_id'],
                $this->MY_RIGHTS,
                $this->MY_RIGHTS_LEVEL
            );
        }
        $sla['main_container'] = $mainContainer;


        /** @var SlaAvailabilityStatusHostsTable $SlaAvailabilityStatusHostsTable */
        $SlaAvailabilityStatusHostsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHosts');

        $slaHostStatusData = $SlaAvailabilityStatusHostsTable->getSlaStatusDataByHostIdAndSlaId($id, $slaId);
        $response = false;
        if (!empty($slaHostStatusData)) {
            $lastEvaluationDate = strtotime(date('d.m.Y', $slaHostStatusData['evaluation_start']->timestamp) . ' 00:00:00');

            $evaluationDateDetails = SlaDaterangesCreator::getDateDetailsForInterval($sla['evaluation_interval'], $lastEvaluationDate);

            $slaHostStatusData['evaluation_total_time_human_readable'] = $UserTime->secondsInHumanShort($slaHostStatusData['evaluation_total_time']);
            $slaHostStatusData['determined_availability_time_human_readable'] = $UserTime->secondsInHumanShort($slaHostStatusData['determined_availability_time']);
            $slaHostStatusData['determined_outage_time_human_readable'] = $UserTime->secondsInHumanShort($slaHostStatusData['determined_outage_time']);
            $slaHostStatusData['evaluation_end'] = $UserTime->format($slaHostStatusData['evaluation_end']);
            $slaHostStatusData['determined_outage_time_percent'] = 0;
            if ($slaHostStatusData['evaluation_total_time']) {
                // Avoid Division by zero
                $slaHostStatusData['determined_outage_time_percent'] = round($slaHostStatusData['determined_outage_time'] / $slaHostStatusData['evaluation_total_time'] * 100, 3);
            }

            $slaHostStatusData['state'] = 'success';
            if ($slaHostStatusData['determined_availability_percent'] < $sla['minimal_availability']) {
                $slaHostStatusData['state'] = 'danger';
            } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaHostStatusData['determined_availability_percent']) {
                $slaHostStatusData['state'] = 'warning';
            }


            /** @var SlaAvailabilityStatusServicesTable $SlaAvailabilityStatusServicesTable */
            $SlaAvailabilityStatusServicesTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServices');
            $slaServicesStatusData = $SlaAvailabilityStatusServicesTable->getSlaStatusDataByHostIdAndSlaId($id, $slaId);

            foreach ($slaServicesStatusData as $key => $slaServiceStatusData) {
                $slaServicesStatusData[$key]['evaluation_total_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['evaluation_total_time']);
                $slaServicesStatusData[$key]['determined_availability_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['determined_availability_time']);
                $slaServicesStatusData[$key]['determined_outage_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['determined_outage_time']);
                $slaServicesStatusData[$key]['evaluation_end'] = $UserTime->format($slaServiceStatusData['evaluation_end']);
                $slaServicesStatusData[$key]['state'] = 'success';
                if ($slaServiceStatusData['determined_availability_percent'] < $sla['minimal_availability']) {
                    $slaServicesStatusData[$key]['state'] = 'danger';
                } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaServiceStatusData['determined_availability_percent']) {
                    $slaServicesStatusData[$key]['state'] = 'warning';
                }
            }

            if (!empty($evaluationDateDetails)) {
                $evaluationDateDetails['start_user_format'] = $UserTime->customFormat('d.m.Y', $evaluationDateDetails['start']);
                $evaluationDateDetails['end_user_format'] = $UserTime->customFormat('d.m.Y', $evaluationDateDetails['end']);
            }

            $response = [
                'host'                  => $slaHostStatusData,
                'services'              => $slaServicesStatusData,
                'evaluationDateDetails' => $evaluationDateDetails
            ];
        }

        $this->set('response', $response);
        $this->set('sla', $sla);
        $this->viewBuilder()->setOption('serialize', ['response', 'sla']);
    }

    /**
     * @param $id
     * @return void
     * @throws \Exception
     */
    public function slaServiceInformation($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template
            return;
        }
        if (!$this->isAngularJsRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var ServicesTable $ServicesTable */
        $ServicesTable = TableRegistry::getTableLocator()->get('Services');

        if (!$ServicesTable->existsById($id)) {
            throw new NotFoundException(__('Invalid service'));
        }

        $service = $ServicesTable->getServiceById($id);

        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        $host = $HostsTable->getHostById($service['host_id']);

        $containerIdsToCheck = Hash::extract($host, 'hosts_to_containers_sharing.{n}.id');
        $containerIdsToCheck[] = $host['container_id'];
        if (!$this->allowedByContainerId($containerIdsToCheck, false)) {
            $this->render403();
            return;
        }


        $User = new User($this->getUser());
        $UserTime = UserTime::fromUser($User);

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        $sla = $SlasTable->get($host['sla_id'], [
            'contain' => [
                'Timeperiods' => 'TimeperiodTimeranges'
            ]
        ]);


        if ($sla === null) {
            $this->set('response', []);
            $this->viewBuilder()->setOption('serialize', ['response']);
            return;
        }


        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        //Load containers information
        if (array_key_exists($sla['container_id'], $this->MY_RIGHTS_LEVEL)) {
            $mainContainer = $ContainersTable->getTreePathForBrowser($host['container_id'], $this->MY_RIGHTS_LEVEL);
        } else {
            //The user only see this host via host sharing
            //We need to "fake" a primary container because the user has no permissions to the real
            //primary container
            $mainContainer = $ContainersTable->getFakePrimaryContainerForHostBrowserDisplay(
                Hash::extract($host, 'hosts_to_containers_sharing.{n}.id'),
                $this->MY_RIGHTS,
                $this->MY_RIGHTS_LEVEL
            );
        }
        $sla['main_container'] = $mainContainer;

        /** @var SlaAvailabilityStatusServicesTable $SlaAvailabilityStatusServicesTable */
        $SlaAvailabilityStatusServicesTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServices');

        $slaServiceStatusData = $SlaAvailabilityStatusServicesTable->getSlaStatusDataByServiceId($id, $host['sla_id']);

        /** @var SlaAvailabilityStatusServicesLogTable $SlaAvailabilityStatusServicesLogTable */
        $SlaAvailabilityStatusServicesLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServicesLog');

        $slaServiceStatusLogData = $SlaAvailabilityStatusServicesLogTable->getSlaStatusLogDataByServiceId($id, $host['sla_id']);
        foreach ($slaServiceStatusLogData as $key => $slaServiceStatusLog) {
            $slaServiceStatusLogData[$key]['evaluation_start'] = $UserTime->customFormat('d.m.Y', $slaServiceStatusLog['evaluation_start']);
            $slaServiceStatusLogData[$key]['state'] = 'success';
            if ($slaServiceStatusLog['determined_availability_percent'] < $sla['minimal_availability']) {
                $slaServiceStatusLogData[$key]['state'] = 'danger';
            } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaServiceStatusLog['determined_availability_percent']) {
                $slaServiceStatusLogData[$key]['state'] = 'warning';
            }
            $slaServiceStatusLogData[$key]['determined_number_outages'] = sizeof($slaServiceStatusLog['sla_service_outages']);
        }
        $slaServiceStatusData['log'] = $slaServiceStatusLogData;
        //debug($slaServiceStatusLogData);
        $response = false;
        if (!empty($slaServiceStatusData)) {
            $lastEvaluationDate = strtotime(date('d.m.Y', $slaServiceStatusData['evaluation_start']->timestamp) . ' 00:00:00');

            $evaluationDateDetails = SlaDaterangesCreator::getDateDetailsForInterval($sla['evaluation_interval'], $lastEvaluationDate);

            $slaServiceStatusData['evaluation_total_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['evaluation_total_time']);
            $slaServiceStatusData['determined_availability_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['determined_availability_time']);
            $slaServiceStatusData['determined_outage_time_human_readable'] = $UserTime->secondsInHumanShort($slaServiceStatusData['determined_outage_time']);
            $slaServiceStatusData['evaluation_end'] = $UserTime->format($slaServiceStatusData['evaluation_end']);
            $slaServiceStatusData['determined_outage_time_percent'] = 0;
            if ($slaServiceStatusData['evaluation_total_time']) {
                // Avoid Division by zero
                $slaServiceStatusData['determined_outage_time_percent'] = round($slaServiceStatusData['determined_outage_time'] / $slaServiceStatusData['evaluation_total_time'] * 100, 3);
            }

            $slaServiceStatusData['state'] = 'success';
            if ($slaServiceStatusData['determined_availability_percent'] < $sla['minimal_availability']) {
                $slaServiceStatusData['state'] = 'danger';
            } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaServiceStatusData['determined_availability_percent']) {
                $slaServiceStatusData['state'] = 'warning';
            }
            if (!empty($evaluationDateDetails)) {
                $evaluationDateDetails['start_user_format'] = $UserTime->customFormat('d.m.Y', $evaluationDateDetails['start']);
                $evaluationDateDetails['end_user_format'] = $UserTime->customFormat('d.m.Y', $evaluationDateDetails['end']);
            }


            $response = [
                'service'               => $slaServiceStatusData,
                'evaluationDateDetails' => $evaluationDateDetails
            ];
        }


        $this->set('response', $response);
        $this->set('sla', $sla);
        $this->viewBuilder()->setOption('serialize', ['response', 'sla']);
    }

    /**
     * @param null $id
     * @throws \Exception
     */
    public function viewDetails($id = null) {
        if (!$this->isApiRequest()) {
            //Only ship template for AngularJs
            return;
        }

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        if (!$SlasTable->existsById($id)) {
            throw new NotFoundException(__('Invalid sla'));
        }

        $sla = $SlasTable->getSlaById($id);

        if (!$this->allowedByContainerId(Hash::extract($sla, 'container_id'))) {
            $this->render403();
            return;
        }

        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');

        //Load containers information
        if (array_key_exists($sla['container_id'], $this->MY_RIGHTS_LEVEL)) {
            $mainContainer = $ContainersTable->getTreePathForBrowser($sla['container_id'], $this->MY_RIGHTS_LEVEL);
        } else {
            //The user only see this host via host sharing
            //We need to "fake" a primary container because the user has no permissions to the real
            //primary container
            $mainContainer = $ContainersTable->getFakePrimaryContainerForHostBrowserDisplay(
                $sla['container_id'],
                $this->MY_RIGHTS,
                $this->MY_RIGHTS_LEVEL
            );
        }
        $sla['main_container'] = $mainContainer;

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        $Sla = new \SLAModule\Lib\Sla($sla);
        $sla['hostsOverview'] = $Sla->getTotalHostsOverview($MY_RIGHTS);

        $User = new User($this->getUser());
        $UserTime = UserTime::fromUser($User);

        /** @var SlaAvailabilityStatusHostsTable $SlaAvailabilityStatusHostsTable */
        $SlaAvailabilityStatusHostsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHosts');

        $GenericFilter = new GenericFilter($this->request);
        $GenericFilter->setFilters([
            'like'  => [
                'Hosts.name'
            ],
            'range' => [
                'determined_availability'
            ]
        ]);

        $PaginateOMat = new PaginateOMat($this, $this->isScrollRequest(), $GenericFilter->getPage());


        $slaDetails = $SlaAvailabilityStatusHostsTable->getSlasStatusDetails(
            $id,
            $GenericFilter,
            $PaginateOMat,
            $MY_RIGHTS
        );


        foreach ($slaDetails as $key => $slaDetail) {
            $slaDetails[$key]['evaluation_total_time_human_readable'] = $UserTime->secondsInHumanShort($slaDetail['evaluation_total_time']);
            $slaDetails[$key]['determined_availability_time_human_readable'] = $UserTime->secondsInHumanShort($slaDetail['determined_availability_time']);
            $slaDetails[$key]['determined_outage_time_human_readable'] = $UserTime->secondsInHumanShort($slaDetail['determined_outage_time']);
            $slaDetails[$key]['evaluation_end'] = $UserTime->format($slaDetail['evaluation_end']);
            $slaDetails[$key]['state'] = 'success';
            if ($slaDetail['determined_availability'] < $sla['minimal_availability']) {
                $slaDetails[$key]['state'] = 'danger';
            } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaDetail['determined_availability']) {
                $slaDetails[$key]['state'] = 'warning';
            }
            foreach ($slaDetail['sla_availability_status_services'] as $serviceKey => $servicedetails) {
                $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['evaluation_total_time_human_readable'] = $UserTime->secondsInHumanShort($servicedetails['evaluation_total_time']);
                $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['determined_availability_time_human_readable'] = $UserTime->secondsInHumanShort($servicedetails['determined_availability_time']);
                $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['determined_outage_time_human_readable'] = $UserTime->secondsInHumanShort($servicedetails['determined_outage_time']);
                $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['evaluation_end'] = $UserTime->format($servicedetails['evaluation_end']);
                $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['state'] = 'success';
                if ($servicedetails['determined_availability_percent'] < $sla['minimal_availability']) {
                    $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['state'] = 'danger';
                } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $servicedetails['determined_availability_percent']) {
                    $slaDetails[$key]['sla_availability_status_services'][$serviceKey]['state'] = 'warning';
                }
            }
        }

        $slaStatus = $SlaAvailabilityStatusHostsTable->getCurrentDeterminedAvailabilityBySlaId($sla['id']);
        $availability = [
            'status_percent' => (float)$slaStatus['determined_availability'],
            'fulfill'        => ($slaStatus['determined_availability'] >= $sla['minimal_availability'])
        ];

        $this->set('sla', $sla);
        $this->set('slaDetails', $slaDetails);
        $this->set('availability', $availability);
        $this->viewBuilder()->setOption('serialize', ['sla', 'slaDetails', 'availability']);
    }

    public function slaAvailabilityOverview() {
        //Only ship HTML template
        return;
    }

    public function slaHostsOverview() {
        //Only ship HTML template
        return;
    }


    /**
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     * @throws \GuzzleHttp\Exception\GuzzleException
     * @throws \Exception
     */
    public function generate() {
        if ($this->isHtmlRequest()) {
            //Only ship HTML template
            return;
        }
        $User = new User($this->getUser());
        $UserTime = $User->getUserTime();
        //User server timezone but with user date format
        $ServerTime = new UserTime(date_default_timezone_get(), $UserTime->getFormatString());

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            /** @var $ContainersTable ContainersTable */
            //$ContainersTable = TableRegistry::getTableLocator()->get('Containers');
            //$MY_RIGHTS = $ContainersTable->resolveChildrenOfContainerIds($this->MY_RIGHTS);
            // ITC-2863 $this->MY_RIGHTS is already resolved and contains all containerIds a user has access to
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        if ($this->request->is('get') && (!$this->isPdfRequest() || !$this->isZipRequest())) {
            $GenericFilter = new GenericFilter($this->request);
            $slas = $SlasTable->getSlasIndex($GenericFilter, null, $MY_RIGHTS);

            foreach ($slas as $index => $sla) {
                $slas[$index]['id'] = (int)($sla['id']);
                $slas[$index]['allowEdit'] = $this->isWritableContainer($sla['container_id']);
            }

            $this->set('slas', $slas);
            $this->viewBuilder()->setOption('serialize', ['slas']);
        }

        if ($this->request->is('post')) {
            $SlaReportForm = new SlaReportForm();
            $SlaReportForm->execute($this->request->getData('Sla', []));

            if (!empty($SlaReportForm->getErrors())) {
                $this->response = $this->response->withStatus(400);
                $this->set('error', $SlaReportForm->getErrors());
                $this->viewBuilder()->setOption('serialize', ['error']);
                return;
            }
            $sla = $SlasTable->get($this->request->getData('Sla.id', -1));

            $reportFormat = $this->request->getData('Sla.format', 'html');
            $hostsBySlaId = $SlasTable->getAllHostsBySlaId($sla->id, $MY_RIGHTS);

            $SlaReport = new SlaReport(
                $this->request->getData('Sla.id', -1),
                $hostsBySlaId,
                strtotime($this->request->getData('Sla.from_date', date('d.m.Y')) . ' 00:00:00'),
                strtotime($this->request->getData('Sla.to_date', date('d.m.Y')) . ' 00:00:00'),
                $ServerTime,
                (int)$this->request->getData('Sla.evaluation', 2)
            );

            $result = $SlaReport->canSlaBeCalculated();
            if ($result === true) {
                if ($reportFormat !== 'html' && $reportFormat !== 'json') {
                    //Report validated successfully.
                    //PDF and CSV reports will get generated by an GET request.

                    $filename = $sla->get('name');
                    $filenameSuffix = '_' . date('Y', strtotime($this->request->getQuery('from_date', date('d.m.Y')) . ' 00:00:00'));;

                    $this->set('success', true);
                    $this->set('filename', $filename . $filenameSuffix . '.' . $reportFormat);
                    $this->viewBuilder()->setOption('serialize', ['success', 'filename']);
                    return;
                }

                /********* HTML report  *********/
                $slaReport = $SlaReport->generateReport();


                $this->set('success', true);
                $this->set('sla', $sla);
                $this->set('report', $slaReport);
                $this->viewBuilder()->setOption('serialize', [
                    'success',
                    'sla',
                    'report'
                ]);
            }

            if (!empty($result['error'])) {
                $this->response = $this->response->withStatus(400);
                $this->set('report_error', $result);
                $this->viewBuilder()->setOption('serialize', ['report_error']);
                return;
            }
        }
        /********* PDF report  *********/
        if ($this->request->is('get') && $this->isPdfRequest()) {
            //Generate PDF report

            $slaId = (int)$this->request->getQuery('Sla.id', -1);
            $hostsBySlaId = $SlasTable->getAllHostsBySlaId($slaId, $MY_RIGHTS);

            $startTimestamp = strtotime($this->request->getQuery('Sla.from_date', date('d.m.Y')) . ' 00:00:00');
            $endTimestamp = strtotime($this->request->getQuery('Sla.to_date', date('d.m.Y')) . ' 00:00:00');

            $User = new User($this->getUser());
            $UserTime = $User->getUserTime();
            $SlaReport = new SlaReport(
                $slaId,
                $hostsBySlaId,
                $startTimestamp,
                $endTimestamp,
                $UserTime,
                (int)$this->request->getQuery('Sla.evaluation', 2)
            );

            $result = $SlaReport->canSlaBeCalculated();
            if ($result === true) {
                //Create report PDF
                $slaReport = $SlaReport->generateReport(true, true);

                $filename = $slaReport['sla']['name'];
                $filenameSuffix = '_' . date('Y', strtotime($this->request->getQuery('from_date', date('d.m.Y')) . ' 00:00:00'));
                $this->set('success', true);
                $this->set('report', $slaReport);
                $this->set('UserTime', $ServerTime);
                $this->set('reportStart', $UserTime->format($startTimestamp));
                $this->set('reportEnd', $UserTime->format($endTimestamp));

                $this->viewBuilder()->setOption(
                    'pdfConfig',
                    [
                        'download' => true,
                        'filename' => $filename . $filenameSuffix . '.pdf'
                    ]
                );
            }
            if (!empty($result['error'])) {
                $this->response = $this->response->withStatus(400);
                $this->set('report_error', $result);
                $this->viewBuilder()->setOption('serialize', ['report_error']);
                return;
            }
        }
        /********* ZIP report  *********/
        if ($this->request->is('get') && $this->isZipRequest()) {
            //Generate report - ZIP with CSV files
            $slaId = (int)$this->request->getQuery('Sla.id', -1);

            $hostsBySlaId = $SlasTable->getAllHostsBySlaId($slaId, $MY_RIGHTS, true);

            $SlaReport = new SlaReport(
                $slaId,
                $hostsBySlaId,
                strtotime($this->request->getQuery('Sla.from_date', date('d.m.Y')) . ' 00:00:00'),
                strtotime($this->request->getQuery('Sla.to_date', date('d.m.Y')) . ' 00:00:00'),
                $ServerTime,
                (int)$this->request->getQuery('Sla.evaluation', 2)
            );
            $result = $SlaReport->canSlaBeCalculated();
            if ($result === true) {
                //Create report CSV
                $slaReport = $SlaReport->generateReport(false);
                $path = !empty($settings['store_path']) ? $settings['store_path'] . DS : ROOT . DS . 'plugins/SLAModule/webroot/tmp/';

                if (!is_dir($path)) {
                    mkdir($path, 0755, true);
                }
                $CsvReport = new CsvReport($slaReport, $path);
                $zipName = $CsvReport->getReportAsZip();

                $this->disableAutoRender();
                $this->set('success', true);
                $this->set('slaReport', $slaReport);
                $this->set('UserTime', $UserTime);

                $this->response = $this->response->withFile(
                    $zipName,
                    [
                        'download' => false,
                    ]
                );
            }
            if (!empty($result['error'])) {
                $this->response = $this->response->withStatus(400);
                $this->set('report_error', $result);
                $this->viewBuilder()->setOption('serialize', ['report_error']);
                return;
            }
        }
    }

    public function slaWidget() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        $config = [
            'Sla' => [
                'id' => 0
            ],
        ];

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var WidgetsTable $WidgetsTable */
        $WidgetsTable = TableRegistry::getTableLocator()->get('Widgets');

        if ($this->request->is('get')) {
            $widgetId = (int)$this->request->getQuery('widgetId', 0);
            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }
            $widget = $WidgetsTable->getWidgetByIdAsCake2($widgetId);

            if ($widget['Widget']['json_data'] !== null && $widget['Widget']['json_data'] !== '') {
                $json = @json_decode($widget['Widget']['json_data'], true);
                $config = Hash::merge($config, $json);
            }

            /** @var SlasTable $SlasTable */
            $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
            $sla = $SlasTable->getSlaById($config['Sla']['id']);
            if (!empty($sla)) {
                $Sla = new \SLAModule\Lib\Sla($sla);
                $sla['hostsAndServicesOverview'] = $Sla->getTotalHostsAndServicesOverview($MY_RIGHTS);
            } else {
                $sla['id'] = null;
            }
            $acl = [
                'sla' => [
                    'viewDetails' => isset($this->PERMISSIONS['slamodule']['slas']['viewDetails']),
                ]
            ];


            $this->set('sla', $sla);
            $this->set('config', $config);
            $this->set('ACL', $acl);

            $this->viewBuilder()->setOption('serialize', [
                'sla',
                'config',
                'ACL'
            ]);
            return;
        }

        if ($this->request->is('post')) {
            $config['Sla'] = $this->request->getData('Sla', []);
            $widgetId = (int)$this->request->getData('Widget.id', 0);

            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }

            $widget = $WidgetsTable->get($widgetId);
            $widget->set('json_data', json_encode($config));

            $WidgetsTable->save($widget);
            if ($widget->hasErrors()) {
                $this->serializeCake4ErrorMessage($widget);
                return;
            } else {
                $this->set('config', $config);
                $this->viewBuilder()->setOption('serialize', ['config']);
                return;
            }
        }
        throw new MethodNotAllowedException();
    }

    /**
     * @throws \Exception
     */
    public function slaCalendarWidget() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        $config = [
            'Sla' => [
                'id' => 0
            ],
        ];

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var WidgetsTable $WidgetsTable */
        $WidgetsTable = TableRegistry::getTableLocator()->get('Widgets');

        if ($this->request->is('get')) {
            $widgetId = (int)$this->request->getQuery('widgetId', 0);
            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }
            $widget = $WidgetsTable->getWidgetByIdAsCake2($widgetId);

            if ($widget['Widget']['json_data'] !== null && $widget['Widget']['json_data'] !== '') {
                $json = @json_decode($widget['Widget']['json_data'], true);
                $config = Hash::merge($config, $json);
            }

            $User = new User($this->getUser());

            $userTimezone = $User->getTimezone();

            if (strlen($userTimezone) < 2) {
                $userTimezone = 'Europe/Berlin';
            }
            $UserTime = new \DateTime($userTimezone);
            $CalendarTime = new CalendarTime();
            $dateDetails = $CalendarTime->getDateDetailsByTimestamp($UserTime->getTimestamp(), true);

            /** @var SlasTable $SlasTable */
            $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
            $sla = $SlasTable->getSlaById($config['Sla']['id']);

            $slaStatusLog = [];
            $slaLastStatus = [];

            if (!empty($sla)) {
                /** @var SlaAvailabilityStatusHostsLogTable $SlaAvailabilityStatusHostsLogTable */
                $SlaAvailabilityStatusHostsLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHostsLog');

                $slaHostsStatusLog = $SlaAvailabilityStatusHostsLogTable->getLowestSlaHostStatusLogBySlaIdWithRangeGroupByEvaluationEnd(
                    $config['Sla']['id'],
                    $dateDetails['start_timestamp'],
                    $dateDetails['end_timestamp'],
                    $MY_RIGHTS
                );

                /** @var SlaAvailabilityStatusServicesLogTable $SlaAvailabilityStatusServicesLogTable */

                $SlaAvailabilityStatusServicesLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServicesLog');

                $slaServicesStatusLog = $SlaAvailabilityStatusServicesLogTable->getLowestSlaServiceStatusLogBySlaIdWithRangeGroupByEvaluationEnd(
                    $config['Sla']['id'],
                    $dateDetails['start_timestamp'],
                    $dateDetails['end_timestamp']
                );


                foreach ($slaHostsStatusLog as $timestamp => $availabilityValue) {
                    $lowestValue = $availabilityValue;
                    if (isset($slaServicesStatusLog[$timestamp]) && $slaServicesStatusLog[$timestamp] < $lowestValue) {
                        $lowestValue = $slaServicesStatusLog[$timestamp];
                    }
                    if ($lowestValue < $sla['minimal_availability']) {
                        $slaClass = 'danger';
                    } else if ($lowestValue > $sla['minimal_availability'] && !empty($sla['warning_threshold']) &&
                        $lowestValue <= $sla['warning_threshold']) {
                        $slaClass = 'warning';
                    } else {
                        $slaClass = 'success';
                    }
                    $slaStatusLog[$timestamp] = [
                        'lowestValue' => $lowestValue,
                        'class'       => $slaClass
                    ];
                }

            } else {
                $sla['id'] = null;
            }

            $acl = [
                'sla' => [
                    'viewDetails' => isset($this->PERMISSIONS['slamodule']['slas']['viewDetails']),
                ]
            ];

            $this->set('dateDetails', $dateDetails);
            $this->set('slaStatusLog', $slaStatusLog);
            $this->set('slaLastStatus', $slaLastStatus);
            $this->set('sla', $sla);
            $this->set('config', $config);
            $this->set('ACL', $acl);

            $this->viewBuilder()->setOption('serialize', [
                'sla',
                'dateDetails',
                'slaStatusLog',
                'slaLastStatus',
                'config',
                'ACL'
            ]);
            return;
        }

        if ($this->request->is('post')) {
            $config['Sla'] = $this->request->getData('Sla', []);
            $widgetId = (int)$this->request->getData('Widget.id', 0);

            /** @var WidgetsTable $WidgetsTable */
            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }

            $widget = $WidgetsTable->get($widgetId);
            $widget->set('json_data', json_encode($config));

            $WidgetsTable->save($widget);
            if ($widget->hasErrors()) {
                $this->serializeCake4ErrorMessage($widget);
                return;
            } else {
                $this->set('config', $config);
                $this->viewBuilder()->setOption('serialize', ['config']);
                return;
            }
        }
        throw new MethodNotAllowedException();
    }

    public function slaSummaryWidget() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }

        $config = [
            'Sla' => [
                'id' => 0
            ],
        ];

        $MY_RIGHTS = [];
        if ($this->hasRootPrivileges === false) {
            $MY_RIGHTS = $this->MY_RIGHTS;
        }

        /** @var WidgetsTable $WidgetsTable */
        $WidgetsTable = TableRegistry::getTableLocator()->get('Widgets');

        if ($this->request->is('get')) {
            $widgetId = (int)$this->request->getQuery('widgetId', 0);
            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }
            $widget = $WidgetsTable->getWidgetByIdAsCake2($widgetId);

            if ($widget['Widget']['json_data'] !== null && $widget['Widget']['json_data'] !== '') {
                $json = @json_decode($widget['Widget']['json_data'], true);
                $config = Hash::merge($config, $json);
            }

            /** @var SlasTable $SlasTable */
            $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
            $sla = $SlasTable->getSlaById($config['Sla']['id']);
            if (!empty($sla)) {
                $Sla = new \SLAModule\Lib\Sla($sla);
                $overview = $Sla->getTotalHostsAndServicesOverviewGroupByContainer($MY_RIGHTS);
                $sla['hostsAndServicesOverview'] = $overview['hostsAndServiceTotalOverview'];
                $sla['hostsCount'] = $overview['hostsCount'];
                $sla['servicesCount'] = $overview['servicesCount'];
            } else {
                $sla['id'] = null;
            }
            $acl = [
                'sla' => [
                    'viewDetails' => isset($this->PERMISSIONS['slamodule']['slas']['viewDetails']),
                ]
            ];


            $this->set('sla', $sla);
            $this->set('config', $config);
            $this->set('ACL', $acl);

            $this->viewBuilder()->setOption('serialize', [
                'sla',
                'config',
                'ACL'
            ]);
            return;
        }

        if ($this->request->is('post')) {
            $config['Sla'] = $this->request->getData('Sla', []);
            $widgetId = (int)$this->request->getData('Widget.id', 0);

            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }

            $widget = $WidgetsTable->get($widgetId);
            $widget->set('json_data', json_encode($config));

            $WidgetsTable->save($widget);
            if ($widget->hasErrors()) {
                $this->serializeCake4ErrorMessage($widget);
                return;
            } else {
                $this->set('config', $config);
                $this->viewBuilder()->setOption('serialize', ['config']);
                return;
            }
        }
        throw new MethodNotAllowedException();
    }

    public function slasSummaryWidget() {
        if (!$this->isApiRequest()) {
            //Only ship HTML template
            return;
        }
        $config = [
            'Slas' => [
                'name' => ''
            ],
        ];


        /** @var WidgetsTable $WidgetsTable */
        $WidgetsTable = TableRegistry::getTableLocator()->get('Widgets');

        if ($this->request->is('get')) {
            $widgetId = (int)$this->request->getQuery('widgetId', 0);
            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }
            $widget = $WidgetsTable->getWidgetByIdAsCake2($widgetId);

            if ($widget['Widget']['json_data'] !== null && $widget['Widget']['json_data'] !== '') {
                $json = @json_decode($widget['Widget']['json_data'], true);
                $config = Hash::merge($config, $json);
            }


            /** @var SlasTable $SlasTable */
            $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
            $slas = $SlasTable->getSlasByContainerIds(
                $this->MY_RIGHTS
            );

            $MY_RIGHTS = [];
            if ($this->hasRootPrivileges === false) {
                $MY_RIGHTS = $this->MY_RIGHTS;
            }

            $totalSlasOverview = [];

            foreach ($slas as $sla) {
                $Sla = new \SLAModule\Lib\Sla($sla);
                $totalSlasOverview[$sla->get('id')]['hostsAndServicesOverview'] = $Sla->getTotalHostsAndServicesOverview($MY_RIGHTS);
            }
            $totalSlasOverview = Api::makeItJavaScriptAble(
                $totalSlasOverview
            );

            $acl = [
                'sla' => [
                    'viewDetails' => isset($this->PERMISSIONS['slamodule']['slas']['viewDetails']),
                ]
            ];


            $this->set('slas', $totalSlasOverview);
            $this->set('config', $config);
            $this->set('ACL', $acl);

            $this->viewBuilder()->setOption('serialize', [
                'slas',
                'config',
                'ACL'
            ]);
            return;
        }

        if ($this->request->is('post')) {
            $config['Sla'] = $this->request->getData('Sla', []);
            $widgetId = (int)$this->request->getData('Widget.id', 0);

            if (!$WidgetsTable->existsById($widgetId)) {
                throw new \RuntimeException('Invalid widget id');
            }

            $widget = $WidgetsTable->get($widgetId);
            $widget->set('json_data', json_encode($config));

            $WidgetsTable->save($widget);
            if ($widget->hasErrors()) {
                $this->serializeCake4ErrorMessage($widget);
                return;
            } else {
                $this->set('config', $config);
                $this->viewBuilder()->setOption('serialize', ['config']);
                return;
            }
        }
        throw new MethodNotAllowedException();
    }

    public function loadSlas() {
        if (!$this->isAngularJsRequest()) {
            throw new MethodNotAllowedException();
        }

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');
        $slas = $SlasTable->getSlasByContainerId(
            $this->MY_RIGHTS,
            'list'
        );
        $slas = Api::makeItJavaScriptAble(
            $slas
        );

        $this->set('slas', $slas);
        $this->viewBuilder()->setOption('serialize', ['slas']);
    }
}
