<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace SLAModule\Lib;


use Cake\Http\Exception\NotFoundException;
use Cake\I18n\FrozenTime;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use itnovum\openITCOCKPIT\Core\Views\UserTime;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsLogTable;
use SLAModule\Model\Table\SlasTable;

class Sla {


    /**
     * @var \SLAModule\Model\Entity\Sla
     */
    private $sla;


    public function __construct(\SLAModule\Model\Entity\Sla $sla) {
        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        if (!$SlasTable->existsById($sla->id)) {
            throw new NotFoundException();
        }
        $this->sla = $sla;
    }

    public function canSlaBeCalculated() {
        $UserTime = new UserTime(date_default_timezone_get(), 'd.m.Y H:i:s');
        if (empty($this->sla['timeperiod']['timeperiod_timeranges'])) {
            return [
                'error'   => true,
                'message' => __(
                    'SLA "{0}" cannot be evaluated. Time evaluation sla data is not available for the selected period "{1}"',
                    $this->sla['name'],
                    $this->sla['timeperiod']['name']
                ),
                'objects' => []
            ];
        }

        if (!empty($this->sla['start_date'])) {
            $now = FrozenTime::now();
            if ($now->timestamp < $this->sla['start_date']->timestamp) {
                return [
                    'error'   => true,
                    'message' => __(
                        'SLA "{0}" cannot be evaluated. The start date of the evaluation is in the future "{1}"',
                        $this->sla['name'],
                        $UserTime->format($this->sla['start_date']->timestamp)
                    ),
                    'objects' => []
                ];
            }
        }

        $this->sla['hosts'] = array_merge(
            $this->sla['hosts'],
            Hash::extract($this->sla, 'hosttemplates.{n}.hosts.{n}')
        );

        if (empty($this->sla['hosts'])) {
            return [
                'error'   => true,
                'message' => __(
                    'SLA "{0}" cannot be calculated. No hosts are available in selected SLA',
                    $this->sla['name']
                ),
                'objects' => []
            ];
        }

        return true;
    }


    /**
     * @throws \Exception
     */
    public function calculateSla() {
        /** @var  SlaAvailabilityStatusHostsLogTable $SlaAvailabilityStatusHostsLogTable */
        $SlaAvailabilityStatusHostsLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHostsLog');
        foreach ($this->sla['hosts'] as $host) {
            $SlaAvailabilityStatusHostsLogTable->calculateAvailabilityForHost($host, $this->sla);

        }
    }

    /**
     * @param $MY_RIGHTS
     * @return array
     */
    public function getTotalHostsOverview($MY_RIGHTS) {
        $hostsTotalOverview = [
            'totalHosts'     => 0,
            'failed'         => [
                'count' => 0,
                'ids'   => []
            ],
            'warning'        => [
                'count' => 0,
                'ids'   => []
            ],
            'passed'         => [
                'count' => 0,
                'ids'   => []
            ],
            'not_calculated' => [
                'count' => 0,
                'ids'   => []
            ],
            'top10'          => []
        ];

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        $allHostsFromSla = $SlasTable->getHostsWithSlaStatusBySlaId($this->sla['id'], $MY_RIGHTS);
        $hostsWithSlaStatusTopTen = [];
        $topTenCounter = 0;

        if (!empty($allHostsFromSla)) {
            $allHostsFromSla = $allHostsFromSla->toArray();
            $hostsTotalOverview['totalHosts'] = sizeof($allHostsFromSla);
            foreach ($allHostsFromSla as $slaHost) {
                if (is_null($slaHost['determined_availability'])) {
                    $hostsTotalOverview['not_calculated']['count']++;
                    $hostsTotalOverview['not_calculated']['ids'][] = $slaHost['id'];
                } else {
                    if ($topTenCounter < 10) {
                        $state = 'success';
                        if ($slaHost['determined_availability'] < $this->sla['minimal_availability']) {
                            $state = 'danger';
                        } else if (!empty($sla['warning_threshold']) && $sla['warning_threshold'] > $slaHost['determined_availability']) {
                            $state = 'warning';
                        }
                        $hostsWithSlaStatusTopTen[] = [
                            'id'                      => $slaHost['id'],
                            'name'                    => $slaHost['name'],
                            'determined_availability' => $slaHost['determined_availability'],
                            'state'                   => $state
                        ];
                        $topTenCounter++;
                    }

                    if ($this->sla['warning_threshold']) {
                        if ($slaHost['determined_availability'] <= $this->sla['warning_threshold']
                            && $slaHost['determined_availability'] > $this->sla['minimal_availability']) {
                            $hostsTotalOverview['warning']['count']++;
                            $hostsTotalOverview['warning']['ids'][] = $slaHost['id'];
                        } else if ($slaHost['determined_availability'] > $this->sla['warning_threshold']) {
                            $hostsTotalOverview['passed']['count']++;
                            $hostsTotalOverview['passed']['ids'][] = $slaHost['id'];
                            continue;
                        }
                    }

                    if ($slaHost['determined_availability'] < $this->sla['minimal_availability']) {
                        $hostsTotalOverview['failed']['count']++;
                        $hostsTotalOverview['failed']['ids'][] = $slaHost['id'];
                    } else {
                        $hostsTotalOverview['passed']['count']++;
                        $hostsTotalOverview['passed']['ids'][] = $slaHost['id'];
                    }
                }
            }
            $hostsTotalOverview['top10'] = $hostsWithSlaStatusTopTen;
        }

        return $hostsTotalOverview;
    }

    public function getTotalHostsAndServicesOverview($MY_RIGHTS) {
        $hostsAndServiceTotalOverview = [
            'hosts'    => [
                'totalHosts'     => 0,
                'failed'         => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'warning'        => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'passed'         => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'not_calculated' => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ]
            ],
            'services' => [
                'totalServices'  => 0,
                'failed'         => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'warning'        => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'passed'         => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ],
                'not_calculated' => [
                    'count'      => 0,
                    'percentage' => 0,
                    'ids'        => []
                ]
            ]
        ];

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        $allHostsFromSla = $SlasTable->getHostsAndServicesWithSlaStatusBySlaId($this->sla['id'], $MY_RIGHTS);

        if (!empty($allHostsFromSla)) {
            $allHostsFromSla = $allHostsFromSla->toArray();
            $hostsAndServiceTotalOverview['hosts']['totalHosts'] = sizeof($allHostsFromSla);
            foreach ($allHostsFromSla as $slaHost) {
                if (is_null($slaHost['sla_availability_status_host']['determined_availability_percent'])) {
                    $hostsAndServiceTotalOverview['hosts']['not_calculated']['count']++;
                    $hostsAndServiceTotalOverview['hosts']['not_calculated']['ids'][] = $slaHost['id'];
                } else {
                    if ($slaHost['sla_availability_status_host']['determined_availability_percent'] < $this->sla['minimal_availability']) {
                        $hostsAndServiceTotalOverview['hosts']['failed']['count']++;
                        $hostsAndServiceTotalOverview['hosts']['failed']['ids'][] = $slaHost['id'];
                    } else if ($this->sla['warning_threshold']) {
                        if ($slaHost['sla_availability_status_host']['determined_availability_percent'] <= $this->sla['warning_threshold']
                            && $slaHost['sla_availability_status_host']['determined_availability_percent'] > $this->sla['minimal_availability']) {
                            $hostsAndServiceTotalOverview['hosts']['warning']['count']++;
                            $hostsAndServiceTotalOverview['hosts']['warning']['ids'][] = $slaHost['id'];
                        } else if ($slaHost['sla_availability_status_host']['determined_availability_percent'] > $this->sla['warning_threshold']) {
                            $hostsAndServiceTotalOverview['hosts']['passed']['count']++;
                            $hostsAndServiceTotalOverview['hosts']['passed']['ids'][] = $slaHost['id'];

                        }
                    } else {
                        $hostsAndServiceTotalOverview['hosts']['passed']['count']++;
                        $hostsAndServiceTotalOverview['hosts']['passed']['ids'][] = $slaHost['id'];
                    }
                }
                $hostsAndServiceTotalOverview['services']['totalServices'] += sizeof($slaHost['services']);
                foreach ($slaHost['services'] as $service) {
                    if (is_null($service['sla_availability_status_service']['determined_availability_percent'])) {
                        $hostsAndServiceTotalOverview['services']['not_calculated']['count']++;
                        $hostsAndServiceTotalOverview['services']['not_calculated']['ids'][] = $service['id'];
                    } else {
                        if ($this->sla['warning_threshold']) {
                            if ($service['sla_availability_status_service']['determined_availability_percent'] <= $this->sla['warning_threshold']
                                && $service['sla_availability_status_service']['determined_availability_percent'] > $this->sla['minimal_availability']) {
                                $hostsAndServiceTotalOverview['services']['warning']['count']++;
                                $hostsAndServiceTotalOverview['services']['warning']['ids'][] = $service['id'];
                            } else if ($service['sla_availability_status_service']['determined_availability_percent'] > $this->sla['warning_threshold']) {
                                $hostsAndServiceTotalOverview['services']['passed']['count']++;
                                $hostsAndServiceTotalOverview['services']['passed']['ids'][] = $service['id'];
                                continue;
                            }
                        }

                        if ($service['sla_availability_status_service']['determined_availability_percent'] < $this->sla['minimal_availability']) {
                            $hostsAndServiceTotalOverview['services']['failed']['count']++;
                            $hostsAndServiceTotalOverview['services']['failed']['ids'][] = $service['id'];
                        } else {
                            $hostsAndServiceTotalOverview['services']['passed']['count']++;
                            $hostsAndServiceTotalOverview['services']['passed']['ids'][] = $service['id'];
                        }
                    }
                }
            }
        }
        if ($hostsAndServiceTotalOverview['hosts']['totalHosts'] > 0) {
            foreach ($hostsAndServiceTotalOverview['hosts'] as $index => $hostOverview) {
                if (isset($hostsAndServiceTotalOverview['hosts'][$index]['count'])) {
                    $hostsAndServiceTotalOverview['hosts'][$index]['percentage'] = sprintf(
                        '%.4f%%',
                        $hostOverview['count'] / $hostsAndServiceTotalOverview['hosts']['totalHosts'] * 100
                    );
                }
            }
        }

        if ($hostsAndServiceTotalOverview['services']['totalServices'] > 0) {
            foreach ($hostsAndServiceTotalOverview['services'] as $index => $hostOverview) {
                if (isset($hostsAndServiceTotalOverview['services'][$index]['count'])) {
                    $hostsAndServiceTotalOverview['services'][$index]['percentage'] = sprintf(
                        '%.4f%%',
                        $hostOverview['count'] / $hostsAndServiceTotalOverview['services']['totalServices'] * 100
                    );
                }
            }
        }

        return $hostsAndServiceTotalOverview;
    }

    public function getTotalHostsAndServicesOverviewGroupByContainer($MY_RIGHTS) {
        $hostsAndServiceTotalOverview = [];
        $hostsCount = 0;
        $servicesCount = 0;
        $hostsOverview = [
            'totalHosts'     => 0,
            'failed'         => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ],
            'passed'         => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ],
            'not_calculated' => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ]

        ];
        $servicesOverview = [
            'totalServices'  => 0,
            'failed'         => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ],
            'passed'         => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ],
            'not_calculated' => [
                'count'      => 0,
                'percentage' => 0,
                'ids'        => []
            ]

        ];

        /** @var SlasTable $SlasTable */
        $SlasTable = TableRegistry::getTableLocator()->get('SLAModule.Slas');

        $allHostsFromSla = $SlasTable->getHostsAndServicesWithSlaStatusBySlaId(
            $this->sla['id'],
            $MY_RIGHTS,
            true
        );

        if (!empty($allHostsFromSla)) {
            $allHostsFromSlaAsArray = $allHostsFromSla->toArray();
            $hostsGroupByContainerId = Hash::combine(
                $allHostsFromSlaAsArray,
                '{n}.id',
                '{n}',
                '{n}.container_id'
            );

            foreach ($hostsGroupByContainerId as $containerId => $hosts) {
                foreach ($hosts as $slaHost) {
                    $hostsCount++;
                    if (!isset($hostsAndServiceTotalOverview[$containerId])) {
                        $hostsAndServiceTotalOverview[$containerId] = [
                            'container' => [
                                'name'      => $slaHost['container']['name'],
                                'full_path' => $slaHost['primary_container']
                            ],
                            'hosts'     => $hostsOverview,
                            'services'  => $servicesOverview
                        ];
                        $hostsAndServiceTotalOverview[$containerId]['hosts']['totalHosts'] = sizeof($hosts);
                    }


                    if (is_null($slaHost['sla_availability_status_host']['determined_availability_percent'])) {
                        $hostsAndServiceTotalOverview[$containerId]['hosts']['not_calculated']['count']++;
                        $hostsAndServiceTotalOverview[$containerId]['hosts']['not_calculated']['ids'][] = $slaHost['id'];
                    } else {
                        if ($slaHost['sla_availability_status_host']['determined_availability_percent'] < $this->sla['minimal_availability']) {
                            $hostsAndServiceTotalOverview[$containerId]['hosts']['failed']['count']++;
                            $hostsAndServiceTotalOverview[$containerId]['hosts']['failed']['ids'][] = $slaHost['id'];
                        } else {
                            $hostsAndServiceTotalOverview[$containerId]['hosts']['passed']['count']++;
                            $hostsAndServiceTotalOverview[$containerId]['hosts']['passed']['ids'][] = $slaHost['id'];
                        }
                    }
                    $hostsAndServiceTotalOverview[$containerId]['services']['totalServices'] += sizeof($slaHost['services']);
                    foreach ($slaHost['services'] as $service) {
                        $servicesCount++;
                        if (is_null($service['sla_availability_status_service']['determined_availability_percent'])) {
                            $hostsAndServiceTotalOverview[$containerId]['services']['not_calculated']['count']++;
                            $hostsAndServiceTotalOverview[$containerId]['services']['not_calculated']['ids'][] = $service['id'];
                        } else {
                            if ($service['sla_availability_status_service']['determined_availability_percent'] < $this->sla['minimal_availability']) {
                                $hostsAndServiceTotalOverview[$containerId]['services']['failed']['count']++;
                                $hostsAndServiceTotalOverview[$containerId]['services']['failed']['ids'][] = $service['id'];
                            } else {
                                $hostsAndServiceTotalOverview[$containerId]['services']['passed']['count']++;
                                $hostsAndServiceTotalOverview[$containerId]['services']['passed']['ids'][] = $service['id'];
                            }
                        }
                    }
                }

                if ($hostsAndServiceTotalOverview[$containerId]['hosts']['totalHosts'] > 0) {
                    foreach ($hostsAndServiceTotalOverview[$containerId]['hosts'] as $index => $hostOverview) {
                        if (isset($hostsAndServiceTotalOverview[$containerId]['hosts'][$index]['count'])) {
                            $hostsAndServiceTotalOverview[$containerId]['hosts'][$index]['percentage'] = sprintf(
                                '%.4f%%',
                                $hostOverview['count'] / $hostsAndServiceTotalOverview[$containerId]['hosts']['totalHosts'] * 100
                            );
                        }
                    }
                }

                if ($hostsAndServiceTotalOverview[$containerId]['services']['totalServices'] > 0) {
                    foreach ($hostsAndServiceTotalOverview[$containerId]['services'] as $index => $hostOverview) {
                        if (isset($hostsAndServiceTotalOverview[$containerId]['services'][$index]['count'])) {
                            $hostsAndServiceTotalOverview[$containerId]['services'][$index]['percentage'] = sprintf(
                                '%.4f%%',
                                $hostOverview['count'] / $hostsAndServiceTotalOverview[$containerId]['services']['totalServices'] * 100
                            );
                        }
                    }
                }
            }
        }

        return [
            'hostsAndServiceTotalOverview' => $hostsAndServiceTotalOverview,
            'hostsCount'                   => $hostsCount,
            'servicesCount'                => $servicesCount,
        ];
    }

}
