<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

namespace SLAModule\Lib;


use App\Lib\Interfaces\HoststatusTableInterface;
use App\Model\Entity\Host;
use App\Model\Table\SystemfailuresTable;
use Cake\ORM\TableRegistry;
use Cake\Utility\Hash;
use itnovum\openITCOCKPIT\Core\DbBackend;
use itnovum\openITCOCKPIT\Core\DowntimeHostConditions;
use itnovum\openITCOCKPIT\Core\Hoststatus;
use itnovum\openITCOCKPIT\Core\HoststatusFields;
use itnovum\openITCOCKPIT\Core\Reports\DaterangesCreator;
use itnovum\openITCOCKPIT\Core\Reports\StatehistoryConverter;
use itnovum\openITCOCKPIT\Core\StatehistoryHostConditions;
use itnovum\openITCOCKPIT\Core\Views\Downtime;
use itnovum\openITCOCKPIT\Core\Views\StatehistoryHost;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsLogTable;
use SLAModule\Model\Table\SlaAvailabilityStatusHostsTable;
use SLAModule\Model\Table\SlaAvailabilityStatusServicesTable;
use SLAModule\Reports\SlaDaterangesCreator;

class SlaHost {
    /**
     * @var \SLAModule\Model\Entity\Sla $sla
     */
    private $Sla;

    /**
     * @var Host
     */
    private $Host;

    /**
     * @var DbBackend
     */
    private $DbBackend;

    /**
     * @param \SLAModule\Model\Entity\Sla $sla
     * @param Host $host
     */
    public function __construct(\SLAModule\Model\Entity\Sla $Sla, Host $Host) {
        $this->Sla = $Sla->toArray();
        $this->Host = $Host;
        $this->DbBackend = new DbBackend();
    }

    /**
     * @return SlaResult
     * @throws \App\Lib\Exceptions\MissingDbBackendException
     */
    public function calculateAvailabilityForHost(): SlaResult {
        /** @var SlaAvailabilityStatusHostsTable $SlaAvailabilityStatusHostsTable */
        $SlaAvailabilityStatusHostsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHosts');

        $firstRun = true;
        $slaStartDate = DaterangesCreator::getInitStartTimeForCalculateAvailability($this->Sla['evaluation_interval']);

        //use start date of sla if defined
        if (!empty($this->Sla['start_date'])) {
            $slaStartDate = $this->Sla['start_date']->timestamp;
        }

        if ($slaStartDate < $this->Host->created->getTimestamp()) {
            // Host was created after SLA start date
            $slaStartDate = $this->Host->created->getTimestamp();
        }

        //use date of last calculated sla if exists
        if (!empty($this->Host['sla_availability_status_host'])) {
            $lastRecordFromLog = $this->Host['sla_availability_status_host'];
            $slaStartDate = (int)$lastRecordFromLog['evaluation_end']->toUnixString();
            $firstRun = false;
        } else {
            // Check the SlaAvailabilityStatusHostsLog Table if records exists
            /** @var SlaAvailabilityStatusHostsLogTable $SlaAvailabilityStatusHostsLogTable */
            $SlaAvailabilityStatusHostsLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHostsLog');

            $query = $SlaAvailabilityStatusHostsLogTable->find()
                ->select([
                    'SlaAvailabilityStatusHostsLog.host_id',
                    'SlaAvailabilityStatusHostsLog.evaluation_end',
                    'SlaAvailabilityStatusHostsLog.evaluation_total_time',
                    'SlaAvailabilityStatusHostsLog.total_time',
                    'SlaAvailabilityStatusHostsLog.determined_outage_time',
                    'SlaAvailabilityStatusHostsLog.determined_availability_percent',
                    'SlaAvailabilityStatusHostsLog.determined_number_outages'

                ])
                ->where([
                    'SlaAvailabilityStatusHostsLog.host_id' => $this->Host->get('id')
                ])
                ->order(['SlaAvailabilityStatusHostsLog.id' => 'desc'])
                ->disableHydration()
                ->limit(1);
            $lastRecordFromLog = $query->first();
            if (!empty($lastRecordFromLog)) {
                $slaStartDate = (int)$lastRecordFromLog['evaluation_end']->toUnixString();
                $firstRun = false;
            }
        }
        $dateRanges = SlaDaterangesCreator::createDateSlicesForSlaByIntervalAndLastUpdateDate(
            $slaStartDate,
            $this->Sla['evaluation_interval'],
            $firstRun
        );
        /** @var SlaAvailabilityStatusHostsLogTable $SlaAvailabilityStatusHostsLogTable */
        $SlaAvailabilityStatusHostsLogTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHostsLog');

        /** @var HoststatusTableInterface $HoststatusTable */

        $HoststatusTable = $this->DbBackend->getHoststatusTable();
        $HoststatusFields = new HoststatusFields($this->DbBackend);
        $HoststatusFields
            ->currentState()
            ->isHardstate()
            ->lastStateChange()
            ->lastHardState()
            ->lastHardStateChange()
            ->maxCheckAttempts()
            ->currentCheckAttempt()
            ->output();
        $hoststatus = $HoststatusTable->byUuid($this->Host->get('uuid'), $HoststatusFields);
        $statehistoryRecords = [];
        if (!empty($hoststatus)) {
            $Hoststatus = new Hoststatus($hoststatus['Hoststatus']);
            $lastStateChange = max($Hoststatus->getLastStateChange(), $Hoststatus->getLastHardStateChange());
            if ($this->Sla['hard_state_only']) {
                $lastStateChange = $Hoststatus->getLastHardStateChange();
            }

            if ($lastStateChange > $slaStartDate) {
                //state change since last sla calculation
                $StatehistoryHostsTable = $this->DbBackend->getStatehistoryHostsTable();
                $Conditions = new StatehistoryHostConditions();
                $Conditions->setFrom($slaStartDate);
                $Conditions->setHostUuid($this->Host->get('uuid'));
                $Conditions->setHardStateTypeAndUpState($this->Sla['hard_state_only']);

                $statehistoryRecords = $StatehistoryHostsTable->getRecordsForReporting($Conditions);
            } else {
                // no state change since last sla calculation
                $statehistoryRecords = [
                    new StatehistoryHost([
                        'hostname'              => $this->Host->get('uuid'),
                        'state_time'            => $lastStateChange,
                        'state_time_usec'       => 0,
                        'state_change'          => false,
                        'state'                 => $Hoststatus->currentState(),
                        'is_hardstate'          => $Hoststatus->isHardState(),
                        'current_check_attempt' => $Hoststatus->getCurrentCheckAttempts(),
                        'max_check_attempts'    => $Hoststatus->getMaxCheckAttempts(),
                        'last_state'            => $Hoststatus->currentState(),
                        'last_hard_state'       => $Hoststatus->getLastHardState(),
                        'output'                => $Hoststatus->getOutput(),
                        'long_output'           => null
                    ])
                ];
            }
        }
        $downtimes = [];
        if ($this->Sla['consider_downtimes']) {
            /** @var SystemfailuresTable $SystemfailuresTable */
            $SystemfailuresTable = TableRegistry::getTableLocator()->get('Systemfailures');

            $systemFailures = $SystemfailuresTable->getSystemfailuresForReportingFormatedAsDowntime(
                $slaStartDate,
                time()
            );

            $DowntimeHostConditions = new DowntimeHostConditions();
            $DowntimeHostConditions->setOrder(['DowntimeHosts.scheduled_start_time' => 'asc']);
            $DowntimeHostConditions->setFrom($slaStartDate);
            $DowntimeHostConditions->setTo(time());
            $DowntimeHostConditions->setHostUuid($this->Host->get('uuid'));

            $DowntimehistoryHostsTable = $this->DbBackend->getDowntimehistoryHostsTable();
            $downtimes = $DowntimehistoryHostsTable->getDowntimesForReporting($DowntimeHostConditions, false, true);
            $formatedDowntimes = [];
            foreach ($downtimes as $downtime) {
                //Use the same format for downtimes and systemdowntimes (also with Statusengine 2 and Statusengine 3)
                $Downtime = new Downtime($downtime);
                $endTime = $Downtime->getScheduledEndTime();
                if ($Downtime->wasCancelled()) {
                    $endTime = $Downtime->getActualEndTime();
                }

                $formatedDowntimes[] = [
                    'author_name'          => $Downtime->getAuthorName(),
                    'comment_data'         => $Downtime->getCommentData(),
                    'entry_time'           => date('Y-m-d H:i:s', $Downtime->getEntryTime()),
                    'scheduled_start_time' => date('Y-m-d H:i:s', $Downtime->getScheduledStartTime()),
                    'scheduled_end_time'   => date('Y-m-d H:i:s', $endTime),
                    'duration'             => $Downtime->getDuration(),
                    'was_started'          => (int)$Downtime->wasStarted(),
                    'internal_downtime_id' => $Downtime->getInternalDowntimeId(),
                    'was_cancelled'        => (int)$Downtime->wasCancelled()
                ];
            }

            $downtimes = Hash::sort(
                Hash::filter(
                    array_merge(
                        $systemFailures['Systemfailure'],
                        $formatedDowntimes
                    )
                ), '{n}.scheduled_start_time', 'ASC'
            );
            if (!empty($downtimes)) {
                $downtimes = DaterangesCreator::mergeTimeOverlapping(
                    array_map(
                        function ($downtimes) {
                            return [
                                'start_time' => strtotime($downtimes['scheduled_start_time']),
                                'end_time'   => strtotime($downtimes['scheduled_end_time']),
                            ];
                        },
                        $downtimes
                    )
                );
            }

        }
        foreach ($dateRanges as $dateRange) {
            $currentStart = new \DateTime(date('d.m.Y H:i:s', $dateRange['start']));
            $i = 0;
            $totalTime = 0;
            $totalOutageTime = 0;
            $availabilityInPercent = 100;
            $totalNumberOfOutages = 0;
            $evaluationTotalTime = $dateRange['end'] - $dateRange['start']; //due to the time change, March has one hour less
            if ($dateRange['entryType'] === 'new') {
                $allTimeSlicesInEvalutionPeriod = SlaDaterangesCreator::createDateRangesForSla(
                    $dateRange['start'],
                    $dateRange['end'],
                    $this->Sla['timeperiod']['timeperiod_timeranges'],
                    $this->Sla['timeperiod']['calendar']
                );

                $evaluationTotalTime = Hash::apply(
                    Hash::map(
                        $allTimeSlicesInEvalutionPeriod, '{n}', [DaterangesCreator::class, 'calculateTotalTime']),
                    '{n}',
                    'array_sum'
                );
            }
            if ($dateRange['entryType'] === 'update' && !empty($lastRecordFromLog)) {
                $evaluationTotalTime = $lastRecordFromLog['evaluation_total_time'];

                $totalTime = $lastRecordFromLog['total_time'];
                $totalOutageTime = $lastRecordFromLog['determined_outage_time'];
                $availabilityInPercent = $lastRecordFromLog['determined_availability_percent'];
                $totalNumberOfOutages = $lastRecordFromLog['determined_number_outages'];
            }
            while ($currentStart->getTimestamp() < $dateRange['end'] && ($currentStart->getTimestamp() + 24 * 3600) < time()) {
                try {
                    $timeSlicesForCurrentDay = $this->getTimeSlices(
                        $currentStart->getTimestamp(),
                        $currentStart->getTimestamp() + 24 * 3600
                    );

                    $timeSlicesForCurrentDay = DaterangesCreator::setDowntimesInTimeslices(
                        $timeSlicesForCurrentDay,
                        $downtimes
                    );
                    $statehistoryRecordsArray = [];
                    foreach ($statehistoryRecords as $statehistoryRecord) {
                        $statehistoryRecordsArray[] = $statehistoryRecord->toArray();
                    }
                    $data = StatehistoryConverter::generateReportDataWithOutages(
                        $timeSlicesForCurrentDay,
                        $statehistoryRecordsArray,
                        $this->Sla['hard_state_only'],
                        true,
                        true
                    );

                    $totalTime += Hash::apply(
                        Hash::map(
                            $timeSlicesForCurrentDay, '{n}', [DaterangesCreator::class, 'calculateTotalTime']),
                        '{n}',
                        'array_sum'
                    );
                    $totalOutageTime += Hash::apply(Hash::map($data['outages'], '{n}', [DaterangesCreator::class, 'calculateTotalTime']), '{n}', 'array_sum');
                    $totalNumberOfOutages += sizeof($data['outages']);
                    $availabilityInPercent = strval(round(100 * ($evaluationTotalTime - $totalOutageTime) / $evaluationTotalTime, 3));

                    $logOutages = [];
                    $data['outages'] = array_merge($data['outages'], $data['outages_summarized_in_downtime']);
                    foreach ($data['outages'] as $outage) {
                        $logOutages[] = [
                            'host_id'      => $this->Host->id,
                            'sla_id'       => $this->Sla['id'],
                            'start_time'   => $outage['start'],
                            'end_time'     => $outage['end'],
                            'output'       => $outage['output'],
                            'is_hardstate' => $outage['is_hardstate'],
                            'in_downtime'  => $outage['is_downtime'] ?? null
                        ];
                    }
                    $slaAvailabilityStatusHostsLog[] = $SlaAvailabilityStatusHostsLogTable->newEntity([
                            'host_id'                         => $this->Host->id,
                            'sla_id'                          => $this->Sla['id'],
                            'evaluation_total_time'           => $evaluationTotalTime,
                            'total_time'                      => $totalTime,
                            'evaluation_start'                => date('Y-m-d H:i:s', $currentStart->getTimestamp()),
                            'evaluation_end'                  => date('Y-m-d H:i:s', $currentStart->getTimestamp() + 24 * 3600),
                            'minimal_availability_time'       => (int)($totalTime * $this->Sla['minimal_availability'] / 100),
                            'determined_availability_time'    => $totalTime - $totalOutageTime,
                            'minimal_availability_percent'    => $this->Sla['minimal_availability'],
                            'determined_availability_percent' => $availabilityInPercent,
                            'determined_outage_time'          => $totalOutageTime,
                            'determined_number_outages'       => $totalNumberOfOutages,
                            'up'                              => $data[0],
                            'down'                            => $data[1],
                            'unreachable'                     => $data[2],
                            'sla_host_outages'                => $logOutages
                        ]
                    );

                } catch (\Exception $e) {
                    //if time slices are empty for example on 8x5 time range
                    $slaAvailabilityStatusHostsLog[] = $SlaAvailabilityStatusHostsLogTable->newEntity([
                        'host_id'                         => $this->Host->id,
                        'sla_id'                          => $this->Sla['id'],
                        'evaluation_total_time'           => $evaluationTotalTime,
                        'total_time'                      => 0,
                        'evaluation_start'                => date('Y-m-d H:i:s', $currentStart->getTimestamp()),
                        'evaluation_end'                  => date('Y-m-d H:i:s', $currentStart->getTimestamp() + 24 * 3600),
                        'minimal_availability_time'       => 0,
                        'determined_availability_time'    => 0,
                        'minimal_availability_percent'    => $this->Sla['minimal_availability'],
                        'determined_availability_percent' => $availabilityInPercent,
                        'determined_outage_time'          => 0,
                        'determined_number_outages'       => $totalNumberOfOutages,
                        'up'                              => 0,
                        'down'                            => 0,
                        'unreachable'                     => 0,
                        'sla_host_outages'                => []
                    ]);
                }
                $currentStart->modify('next day');
                $i++;
            }
        }
        $errors = [];
        $statusErrors = [];
        if (!empty($slaAvailabilityStatusHostsLog)) {
            if ($SlaAvailabilityStatusHostsLogTable->saveMany($slaAvailabilityStatusHostsLog)) {
                //successfully finished
                //save last log data in sla host status table

                $lastLogEntry = $slaAvailabilityStatusHostsLog[sizeof($slaAvailabilityStatusHostsLog) - 1];
                $lastLogArray = $lastLogEntry->toArray();
                foreach (['id', 'created', 'updated'] as $unwantedKey) {
                    if (isset($lastLogArray[$unwantedKey])) {
                        unset($lastLogArray[$unwantedKey]);
                    }
                }

                if ($this->Host->get('sla_availability_status_host') !== null) {
                    $entity = $this->Host->get('sla_availability_status_host');

                    $entity = $SlaAvailabilityStatusHostsTable->patchEntity($entity, $lastLogArray);
                } else {
                    $entity = $SlaAvailabilityStatusHostsTable->newEntity($lastLogArray);
                }


                $SlaAvailabilityStatusHostsTable->save($entity);
                if ($entity->hasErrors()) {
                    $statusErrors = $entity->getErrors();
                }

                return new SlaResult(empty($statusErrors), [], $statusErrors);
            }

            foreach ($slaAvailabilityStatusHostsLog as $entity) {
                $errors[] = $entity->getErrors();
            }
        }

        $success = empty($errors) && empty($statusErrors);
        return new SlaResult($success, $errors, $statusErrors);
    }

    /**
     * @param $start
     * @param $end
     * @return array
     * @throws \Exception
     */
    private function getTimeSlices($start, $end) {
        $timeSlicesGlobal = Hash::insert(
            SlaDaterangesCreator::createDateRangesForSlaForOneDay(
                $start,
                $this->Sla['timeperiod']['timeperiod_timeranges']
            ),
            '{n}.is_downtime', false
        );

        $timeSlicesCounter = sizeof($timeSlicesGlobal);

        if (!empty($timeSlicesGlobal)) {
            if ($timeSlicesGlobal[0]['start'] < $start) {
                $timeSlicesGlobal[0]['start'] = $start;
            }
            if ($timeSlicesGlobal[$timeSlicesCounter - 1]['end'] > $end) {
                $timeSlicesGlobal[$timeSlicesCounter - 1]['end'] = $end;
            }
        }

        if ($this->Sla['timeperiod']['calendar_id'] === 0) {
            return $timeSlicesGlobal;
        }
        //filter holidays if needed
        if (!empty($this->Sla['timeperiod']['calendar']['calendar_holidays'])) {
            $holidays = Hash::extract($this->Sla['timeperiod']['calendar']['calendar_holidays'], '{n}.date');
            foreach ($timeSlicesGlobal as $key => $timeSliceGlobal) {
                if (in_array(date('Y-m-d', $timeSliceGlobal['start']), $holidays, true)) {
                    unset($timeSlicesGlobal[$key]);
                }
            }
            if (empty($timeSlicesGlobal)) {
                throw new \Exception(__(
                    'Time evaluation sla data is not available for the selected period ({0} with calendar: {1})',
                    $this->Sla['timeperiod']['name'],
                    $this->Sla['timeperiod']['calendar']['name']
                ));
            }
        }

        return $timeSlicesGlobal;
    }

    public function updateSlaServiceStatus() {
        /** @var SlaAvailabilityStatusHostsTable $SlaAvailabilityStatusHostsTable */
        $SlaAvailabilityStatusHostsTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusHosts');
        $currentSlaHoststatus = $SlaAvailabilityStatusHostsTable->getCurrentSlaStatusByHostIdAndSlaId(
            $this->Host->id,
            $this->Sla['id']
        );
        if (!empty($currentSlaHoststatus)) {
            /** @var SlaAvailabilityStatusServicesTable $SlaAvailabilityStatusServicesTable */
            $SlaAvailabilityStatusServicesTable = TableRegistry::getTableLocator()->get('SLAModule.SlaAvailabilityStatusServices');
            $slaMinimalAvailabilityServiceStatus = $SlaAvailabilityStatusServicesTable->getServiceWithMinimalDeterminedAvailabilityByHostIdAndSlaId(
                $this->Host->id,
                $this->Sla['id'],
                $currentSlaHoststatus['evaluation_start']
            );
            if (isset($slaMinimalAvailabilityServiceStatus['min_determined_services_availability_percent'])) {
                $currentSlaHoststatus->set(
                    'lowest_determined_availability_of_services_percent',
                    $slaMinimalAvailabilityServiceStatus['min_determined_services_availability_percent']
                );
                $SlaAvailabilityStatusHostsTable->save($currentSlaHoststatus);
            }
        }
    }


    private function debugDateRanges(array $dateRanges) {
        echo PHP_EOL;
        foreach ($dateRanges as $range) {
            if (isset($range['entryType'])) {
                // Timerange
                printf('Start: %s  End: %s  entryType: %s%s',
                    date('l d.m.Y H:i:s', $range['start']),
                    date('l d.m.Y H:i:s', $range['end']),
                    $range['entryType'],
                    PHP_EOL
                );
            } else {
                // Timeslice (24x7, 8x5)
                printf('Start: %s  End: %s %s',
                    date('l d.m.Y H:i:s', $range['start']),
                    date('l d.m.Y H:i:s', $range['end']),
                    PHP_EOL
                );
            }
        }
    }
}
