<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostsTable;
use Cake\Database\Expression\Comparison;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\ResultSetInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Core\HostConditions;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use SLAModule\Model\Entity\HostMappingRule;

/**
 * HostMappingRules Model
 *
 * @property SlasTable&BelongsTo $Slas
 *
 * @method HostMappingRule newEmptyEntity()
 * @method HostMappingRule newEntity(array $data, array $options = [])
 * @method HostMappingRule[] newEntities(array $data, array $options = [])
 * @method HostMappingRule get($primaryKey, $options = [])
 * @method HostMappingRule findOrCreate($search, ?callable $callback = null, $options = [])
 * @method HostMappingRule patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method HostMappingRule[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method HostMappingRule|false save(EntityInterface $entity, $options = [])
 * @method HostMappingRule saveOrFail(EntityInterface $entity, $options = [])
 * @method HostMappingRule[]|ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method HostMappingRule[]|ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method HostMappingRule[]|ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method HostMappingRule[]|ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class HostMappingRulesTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('sla_host_mapping_rules');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Slas', [
            'foreignKey' => 'sla_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.Slas',
        ]);
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('sla_id')
            ->notEmptyString('sla_id');

        $validator
            ->scalar('host_keywords')
            ->maxLength('host_keywords', 255)
            ->allowEmptyString('host_keywords', __('Please select at least one option'), function ($context) {
                if ($context['data']['save_options'] === false) {
                    return true;
                }
                return !empty($context['data']['host_not_keywords']) ||
                    !empty($context['data']['hostname_regex']);
            });

        $validator
            ->scalar('host_not_keywords')
            ->maxLength('host_not_keywords', 255)
            ->allowEmptyString('host_not_keywords', __('Please select at least one option'), function ($context) {
                if ($context['data']['save_options'] === false) {
                    return true;
                }
                return !empty($context['data']['host_keywords']) ||
                    !empty($context['data']['hostname_regex']);
            });

        $validator
            ->scalar('hostname_regex')
            ->maxLength('hostname_regex', 2500)
            ->requirePresence('hostname_regex', 'create')
            ->allowEmptyString('hostname_regex', __('Please select at least one option'), function ($context) {
                if ($context['data']['save_options'] === false) {
                    return true;
                }
                return !empty($context['data']['host_keywords']) ||
                    !empty($context['data']['host_not_keywords']);
            })
            ->add('hostname_regex', 'custom', [
                'rule'    => [$this, 'isRegularExpression'],
                'message' => __('Your host name regex is an invalid regular expression.')
            ]);
        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('sla_id', 'Slas'), ['errorField' => 'sla_id']);

        return $rules;
    }

    public function isRegularExpression($string) {
        return @preg_match('`' . $string . '`', '') !== false;
    }

    /**
     * @param mixed $value
     * @param array $context
     * @return bool
     *
     * Custom validation rule for host_keywords and/or host_not_keywords and/or hostname_regex fields.
     */
    public function atLeastOne($value, $context) {
        return !empty($context['data']['host_keywords']) || !empty($context['data']['host_not_keywords']) ||
            !empty($context['data']['hostname_regex']);
    }

    /**
     * @param $slaId
     * @return bool
     */
    public function existsBySlaId($slaId) {
        return $this->exists(['HostMappingRules.sla_id' => $slaId]);
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param HostConditions $HostConditions
     * @param bool $onlyUnassigned
     * @param $type
     * @param $PaginateOMat
     * @return void
     */
    public function getHostsIndex(GenericFilter $GenericFilter, HostConditions $HostConditions, bool $onlyUnassigned = false, $type = 'all', $PaginateOMat = null) {
        //Lookup for the tenant container of $container_id
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $MY_RIGHTS = $HostConditions->getContainerIds();
        if ($type === 'list') {
            //Get hosts with host template sla_id
            $query = $HostsTable->find('list', [
                'keyField'   => 'id',
                'valueField' => function ($row) {
                    return $row['hosttemplate']['sla_id'];
                }
            ]);
        } else {
            $query = $HostsTable->find();
            $query->select([
                'Hosts.id',
                'Hosts.name',
                'Hosts.address',
                'Hosts.sla_id',
                'hostdescription' => $query->newExpr('IF(Hosts.description IS NULL, Hosttemplates.description, Hosts.description)'),
                'hosttags'        => $query->newExpr('IF(Hosts.tags IS NULL, Hosttemplates.tags, Hosts.tags)'),
                'container'       => $query->newExpr('SELECT Containers.name')
            ]);
            if (!empty($MY_RIGHTS)) {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                AND containers_temp.id IN (' . implode(',', $MY_RIGHTS) . ')
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
                ]);
            } else {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
                ]);
            }
            $query->innerJoinWith('Containers');
        }

        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }

        $query->contain([
            'HostsToContainersSharing',
            'Hosttemplates' => [
                'fields' => [
                    'Hosttemplates.id',
                    'Hosttemplates.sla_id'
                ]
            ]
        ]);


        $where = $GenericFilter->genericFilters();
        $where['Hosts.disabled'] = (int)$HostConditions->includeDisabled();

        if (isset($where['Hosts.name rlike'])) {
            if ($this->isRegularExpression($where['Hosts.name rlike'])) {
                $where[] = new Comparison(
                    'Hosts.name',
                    $where['Hosts.name rlike'],
                    'string',
                    'RLIKE'
                );
            }
            unset($where['Hosts.name rlike']);
        }

        if (isset($where['hostdescription LIKE'])) {
            $where[] = new Comparison(
                'IF((Hosts.description IS NULL OR Hosts.description=""), Hosttemplates.description, Hosts.description)',
                $where['hostdescription LIKE'],
                'string',
                'LIKE'
            );
            unset($where['hostdescription LIKE']);
        }

        if (isset($where['Hosts.keywords rlike'])) {
            $where[] = new Comparison(
                'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
                $where['Hosts.keywords rlike'],
                'string',
                'RLIKE'
            );
            unset($where['Hosts.keywords rlike']);
        }

        if (isset($where['Hosts.not_keywords not rlike'])) {
            $where[] = new Comparison(
                'IF((Hosts.tags IS NULL OR Hosts.tags=""), Hosttemplates.tags, Hosts.tags)',
                $where['Hosts.not_keywords not rlike'],
                'string',
                'NOT RLIKE'
            );
            unset($where['Hosts.not_keywords not rlike']);
        }

        //get only unassigned hosts
        if ($onlyUnassigned) {
            $where[] = 'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id) IS NULL';
        }

        $query->where($where);
        $query->disableHydration();
        $query->group(['Hosts.id']);

        if ($type === 'list') {
            if (empty($query)) {
                return [];
            }
            return $query->toArray();
        }

        $query->order(
            array_merge(
                $GenericFilter->getOrderForPaginator('Hosts.name', 'desc'),
                ['Hosts.id' => 'asc']
            )
        );

        if ($PaginateOMat === null) {
            //Just execute query
            if (empty($query)) {
                return [];
            }
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }

    /**
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getHostMappingRuleBySlaId($slaId) {
        return $this->find()
            ->where([
                'HostMappingRules.sla_id' => $slaId
            ])
            ->first();
    }
}
