<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.


declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostsTable;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use SLAModule\Lib\SlaStatusLogConditions;
use SLAModule\Model\Entity\SlaAvailabilityStatusHostLog;

/**
 * SlaAvailabilityStatusHostsLog Model
 *
 * @property HostsTable&BelongsTo $Hosts
 * @property SlasTable&BelongsTo $Slas
 *
 * @method SlaAvailabilityStatusHostLog newEmptyEntity()
 * @method SlaAvailabilityStatusHostLog newEntity(array $data, array $options = [])
 * @method SlaAvailabilityStatusHostLog[] newEntities(array $data, array $options = [])
 * @method SlaAvailabilityStatusHostLog get($primaryKey, $options = [])
 * @method SlaAvailabilityStatusHostLog findOrCreate($search, ?callable $callback = null, $options = [])
 * @method SlaAvailabilityStatusHostLog patchEntity(\Cake\Datasource\EntityInterface $entity, array $data, array $options = [])
 * @method SlaAvailabilityStatusHostLog[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method SlaAvailabilityStatusHostLog|false save(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusHostLog saveOrFail(\Cake\Datasource\EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusHostLog[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHostLog[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHostLog[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHostLog[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class SlaAvailabilityStatusHostsLogTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('sla_availability_status_hosts_log');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts',
        ]);
        $this->belongsTo('Slas', [
            'foreignKey' => 'sla_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.Slas',
        ]);

        $this->hasMany('SlaHostOutages');
    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosts':
                $coreTable->hasMany('SlaAvailabilityStatusHostsLog', [
                    'className'        => 'SLAModule.SlaAvailabilityStatusHostsLog',
                    'dependent'        => true,
                    'foreignKey'       => 'host_id',
                    'cascadeCallbacks' => true
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('host_id')
            ->notEmptyString('host_id');

        $validator
            ->integer('sla_id')
            ->notEmptyString('sla_id');

        $validator
            ->numeric('minimal_availability_percent')
            ->requirePresence('minimal_availability_percent', 'create')
            ->notEmptyString('minimal_availability_percent');

        $validator
            ->numeric('determined_availability_percent')
            ->requirePresence('determined_availability_percent', 'create')
            ->notEmptyString('determined_availability_percent');

        $validator
            ->integer('total_time')
            ->requirePresence('total_time', 'create')
            ->notEmptyString('total_time');

        $validator
            ->integer('minimal_availability_time')
            ->requirePresence('minimal_availability_time', 'create')
            ->notEmptyString('minimal_availability_time');

        $validator
            ->integer('determined_availability_time')
            ->requirePresence('determined_availability_time', 'create')
            ->notEmptyString('determined_availability_time');

        $validator
            ->integer('determined_outage_time')
            ->requirePresence('determined_outage_time', 'create')
            ->notEmptyString('determined_outage_time');

        $validator
            ->integer('determined_number_outages')
            ->notEmptyString('determined_number_outages');

        $validator
            ->dateTime('evaluation_start')
            ->requirePresence('evaluation_start', 'create')
            ->notEmptyDateTime('evaluation_start');

        $validator
            ->dateTime('evaluation_end')
            ->requirePresence('evaluation_end', 'create')
            ->notEmptyDateTime('evaluation_end');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('host_id', 'Hosts'), ['errorField' => 'host_id']);
        $rules->add($rules->existsIn('sla_id', 'Slas'), ['errorField' => 'sla_id']);

        return $rules;
    }

    /**
     * @param SlaStatusLogConditions $SlaStatusLogConditions
     * @param $PaginateOMat
     * @param $enableHydration
     * @return array
     */
    public function getSlaHostStatusLogIndex(SlaStatusLogConditions $SlaStatusLogConditions, $PaginateOMat = null, $enableHydration = true) {
        $query = $this->find()
            ->contain(['SlaHostOutages'])
            ->where([
                'SlaAvailabilityStatusHostsLog.host_id'            => $SlaStatusLogConditions->getId(),
                'SlaAvailabilityStatusHostsLog.evaluation_start >' => date('Y-m-d H:i:s', $SlaStatusLogConditions->getFrom()),
                'SlaAvailabilityStatusHostsLog.evaluation_start <' => date('Y-m-d H:i:s', $SlaStatusLogConditions->getTo())
            ])
            ->order($SlaStatusLogConditions->getOrder());

        if ($SlaStatusLogConditions->hasConditions()) {
            $query->andWhere($SlaStatusLogConditions->getConditions());
        }


        $query->enableHydration($enableHydration);

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param $hostId
     * @param $from
     * @param $to
     * @return array
     */
    public function getSlaHostStatusLogByHostIdWithRange($hostId, $from, $to): array {
        $query = $this->find();
        $query->contain([
            'SlaHostOutages'
        ]);

        $query->where([
            'SlaAvailabilityStatusHostsLog.host_id'             => $hostId,
            'SlaAvailabilityStatusHostsLog.evaluation_start >=' => date('Y-m-d H:i:s', $from),
            'SlaAvailabilityStatusHostsLog.evaluation_end <='   => date('Y-m-d H:i:s', $to)
        ])
            ->order([
                'SlaAvailabilityStatusHostsLog.evaluation_start' => 'asc',
                'SlaAvailabilityStatusHostsLog.id'
            ])
            ->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }

    /**
     * @param $slaId
     * @param $from
     * @param $to
     * @return array
     */
    public function getLowestSlaHostStatusLogBySlaIdWithRangeGroupByEvaluationEnd($slaId, $from, $to, $MY_RIGHTS = []): array {
        //$query = $this->find();

        $query = $this->find('list', [
            'keyField'   => function ($row) {
                return $row['evaluation_start']->timestamp;
            },
            'valueField' => 'determined_availability_host'
        ]);

        $query->select([
            'SlaAvailabilityStatusHostsLog.evaluation_start',
            'determined_availability_host' => $query->func()
                ->min('determined_availability_percent')
        ]);

        if (!empty($MY_RIGHTS)) {
            $query
                ->innerJoinWith('Hosts')
                ->innerJoinWith('Hosts.HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                    $q->where([
                        'HostsToContainersSharing.id IN ' => $MY_RIGHTS
                    ]);

                    return $q;
                });
        }
        $query->where([
            'SlaAvailabilityStatusHostsLog.sla_id'              => $slaId,
            'SlaAvailabilityStatusHostsLog.evaluation_start >=' => date('Y-m-d H:i:s', $from),
            'SlaAvailabilityStatusHostsLog.evaluation_start <='   => date('Y-m-d H:i:s', $to)
        ])
            ->order([
                'SlaAvailabilityStatusHostsLog.evaluation_start' => 'asc',
                'SlaAvailabilityStatusHostsLog.id'
            ])
            ->group('SlaAvailabilityStatusHostsLog.evaluation_end');
        $query->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }
}
