<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.


declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Lib\Traits\PluginManagerTableTrait;
use App\Model\Table\HostsTable;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use SLAModule\Model\Entity\SlaAvailabilityStatusHost;

/**
 * SlaAvailabilityStatusHosts Model
 *
 * @property HostsTable&BelongsTo $Hosts
 * @property SlasTable&BelongsTo $Slas
 *
 * @method SlaAvailabilityStatusHost newEmptyEntity()
 * @method SlaAvailabilityStatusHost newEntity(array $data, array $options = [])
 * @method SlaAvailabilityStatusHost[] newEntities(array $data, array $options = [])
 * @method SlaAvailabilityStatusHost get($primaryKey, $options = [])
 * @method SlaAvailabilityStatusHost findOrCreate($search, ?callable $callback = null, $options = [])
 * @method SlaAvailabilityStatusHost patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method SlaAvailabilityStatusHost[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method SlaAvailabilityStatusHost|false save(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusHost saveOrFail(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusHost[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHost[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHost[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusHost[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class SlaAvailabilityStatusHostsTable extends Table {

    use PluginManagerTableTrait;
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('sla_availability_status_hosts');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts',
        ]);

        $this->belongsTo('Slas', [
            'foreignKey' => 'sla_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.Slas',
        ]);

        $this->hasMany('SlaAvailabilityStatusServices', [
            'foreignKey' => 'host_id',
            'bindingKey' => 'host_id', // TblOrganisationSubstances.app_id
            'className'  => 'SLAModule.SlaAvailabilityStatusServices',
        ]);
    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosts':
                $coreTable->hasOne('SlaAvailabilityStatusHosts', [
                    'className'        => 'SLAModule.SlaAvailabilityStatusHosts',
                    'dependent'        => true,
                    'foreignKey'       => 'host_id',
                    'cascadeCallbacks' => true
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('host_id')
            ->notEmptyString('host_id')
            ->add('host_id', 'unique', ['rule' => 'validateUnique', 'provider' => 'table']);

        $validator
            ->integer('sla_id')
            ->notEmptyString('sla_id');

        $validator
            ->numeric('minimal_availability_percent')
            ->requirePresence('minimal_availability_percent', 'create')
            ->notEmptyString('minimal_availability_percent');

        $validator
            ->numeric('determined_availability_percent')
            ->requirePresence('determined_availability_percent', 'create')
            ->notEmptyString('determined_availability_percent');

        $validator
            ->integer('total_time')
            ->requirePresence('total_time', 'create')
            ->notEmptyString('total_time');

        $validator
            ->integer('minimal_availability_time')
            ->requirePresence('minimal_availability_time', 'create')
            ->notEmptyString('minimal_availability_time');

        $validator
            ->integer('determined_availability_time')
            ->requirePresence('determined_availability_time', 'create')
            ->notEmptyString('determined_availability_time');

        $validator
            ->integer('determined_outage_time')
            ->requirePresence('determined_outage_time', 'create')
            ->notEmptyString('determined_outage_time');

        $validator
            ->integer('determined_number_outages')
            ->notEmptyString('determined_number_outages');

        $validator
            ->dateTime('evaluation_start')
            ->requirePresence('evaluation_start', 'create')
            ->notEmptyDateTime('evaluation_start');

        $validator
            ->dateTime('evaluation_end')
            ->requirePresence('evaluation_end', 'create')
            ->notEmptyDateTime('evaluation_end');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('host_id', 'Hosts'), ['errorField' => 'host_id']);
        $rules->add($rules->existsIn('sla_id', 'Slas'), ['errorField' => 'sla_id']);

        return $rules;
    }

    /**
     * @param $hostId
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getSlaStatusDataByHostIdAndSlaId($hostId, $slaId) {
        $query = $this->find()
            ->contain([
                'Hosts' => function (Query $query) use ($hostId, $slaId) {
                    return $query
                        ->select([
                            'Hosts.id',
                            'Hosts.name'
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Hosts.id'       => $hostId,
                            'Hosts.disabled' => 0
                        ]);
                }
            ])
            ->where([
                'SlaAvailabilityStatusHosts.sla_id' => $slaId
            ])
            ->disableHydration();
        return $query->first();
    }

    /**
     * @param $hostId
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getCurrentSlaStatusByHostIdAndSlaId($hostId, $slaId) {
        $query = $this->find()
            ->select([
                'SlaAvailabilityStatusHosts.id',
                'SlaAvailabilityStatusHosts.evaluation_start',
            ])
            ->where([
                'SlaAvailabilityStatusHosts.host_id' => $hostId,
                'SlaAvailabilityStatusHosts.sla_id'  => $slaId
            ])
            ->disableAutoFields()
            ->disableResultsCasting();
        return $query->first();
    }

    /**
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getCurrentDeterminedAvailabilityBySlaId($slaId) {
        $query = $this->find();

        $query->select([
            'determined_availability' => $query->func()
                ->min(
                    $query->newExpr(
                        'IF(SlaAvailabilityStatusHosts.determined_availability_percent > 
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent,  
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent, 
                    SlaAvailabilityStatusHosts.determined_availability_percent)'
                    )
                )
        ])
            ->where([
                'SlaAvailabilityStatusHosts.sla_id' => $slaId
            ])
            ->disableAutoFields()
            ->disableResultsCasting();
        return $query->first();
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param $MY_RIGHTS
     * @return array
     */
    public function getSlasStatusDetails($slaId, GenericFilter $GenericFilter, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find();

        $query->select([
            'SlaAvailabilityStatusHosts.host_id',
            'SlaAvailabilityStatusHosts.evaluation_total_time',
            'SlaAvailabilityStatusHosts.determined_outage_time',
            'SlaAvailabilityStatusHosts.evaluation_end',
            'SlaAvailabilityStatusHosts.determined_availability_percent',
            'SlaAvailabilityStatusHosts.evaluation_total_time',
            'SlaAvailabilityStatusHosts.determined_availability_time',
            'SlaAvailabilityStatusHosts.determined_outage_time',
            'SlaAvailabilityStatusHosts.evaluation_end',
            'determined_availability' =>
                $query->newExpr(
                    'IF(SlaAvailabilityStatusHosts.determined_availability_percent > 
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent,  
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent, 
                    SlaAvailabilityStatusHosts.determined_availability_percent)'
                )

        ]);

        $query->contain([
            'Hosts' => function (Query $query) {
                return $query
                    ->select([
                        'Hosts.id',
                        'Hosts.name',
                        'Hosts.description'
                    ])
                    ->disableAutoFields()
                    ->where([
                        'Hosts.disabled' => 0
                    ]);
            },

            'SlaAvailabilityStatusServices' => function (Query $query) {
                return $query
                    ->contain([
                        'Services' => function (Query $query) {
                            $query = $query->contain([
                                'Servicetemplates' => function (Query $query) {
                                    return $query->select([
                                        'Servicetemplates.id'
                                    ])->disableAutoFields();
                                }
                            ]);
                            return $query
                                ->select([
                                    'Services.id',
                                    'servicename' => $query->newExpr('IF(Services.name IS NULL, Servicetemplates.name, Services.name)')
                                ])
                                ->disableAutoFields()
                                ->where([
                                    'Services.disabled' => 0
                                ]);

                        }
                    ])->order([
                        'SlaAvailabilityStatusServices.determined_availability_percent' => 'ASC'
                    ]);
            }
        ]);
        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }
        $query->where([
            'SlaAvailabilityStatusHosts.sla_id' => $slaId
        ])
            ->disableResultsCasting()
            ->disableHydration();

        $where = $GenericFilter->genericFilters();
        $having = [];
        if (isset($where['determined_availability >='])) {
            $having['determined_availability >='] = $where['determined_availability >='];
            unset($where['determined_availability >=']);
        }

        if (isset($where['determined_availability <='])) {
            $having['determined_availability <='] = $where['determined_availability <='];
            unset($where['determined_availability <=']);
        }
        if (!empty($where)) {
            $query->where($where);
        }

        if (!empty($having)) {
            $query->having($having);
        }

        $query->order($GenericFilter->getOrderForPaginator('determined_availability', 'asc'));
        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }
}
