<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.


declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use SLAModule\Lib\SlaStatusLogConditions;
use SLAModule\Model\Entity\SlaAvailabilityStatusServiceLog;

/**
 * SlaAvailabilityStatusLogServices Model
 *
 * @property HostsTable&BelongsTo $Hosts
 * @property ServicesTable&BelongsTo $Services
 * @property SlasTable&BelongsTo $Slas
 *
 * @method SlaAvailabilityStatusServiceLog newEmptyEntity()
 * @method SlaAvailabilityStatusServiceLog newEntity(array $data, array $options = [])
 * @method SlaAvailabilityStatusServiceLog[] newEntities(array $data, array $options = [])
 * @method SlaAvailabilityStatusServiceLog get($primaryKey, $options = [])
 * @method SlaAvailabilityStatusServiceLog findOrCreate($search, ?callable $callback = null, $options = [])
 * @method SlaAvailabilityStatusServiceLog patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method SlaAvailabilityStatusServiceLog[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method SlaAvailabilityStatusServiceLog|false save(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusServiceLog saveOrFail(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusServiceLog[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusServiceLog[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusServiceLog[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusServiceLog[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin \Cake\ORM\Behavior\TimestampBehavior
 */
class SlaAvailabilityStatusServicesLogTable extends Table {
    use PaginationAndScrollIndexTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('sla_availability_status_services_log');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts',
        ]);
        $this->belongsTo('Services', [
            'foreignKey' => 'service_id',
            'joinType'   => 'INNER',
            'className'  => 'Services',
        ]);
        $this->belongsTo('Slas', [
            'foreignKey' => 'sla_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.Slas',
        ]);

        $this->hasMany('SlaServiceOutages');
    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Services':
                $coreTable->hasMany('SlaAvailabilityStatusServicesLog', [
                    'className'        => 'SLAModule.SlaAvailabilityStatusServicesLog',
                    'dependent'        => true,
                    'foreignKey'       => 'service_id',
                    'cascadeCallbacks' => true
                ]);
                break;
        }
    }

    /**
     * Default validation rules.
     *
     * @param \Cake\Validation\Validator $validator Validator instance.
     * @return \Cake\Validation\Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('host_id')
            ->notEmptyString('host_id');

        $validator
            ->integer('service_id')
            ->notEmptyString('service_id');

        $validator
            ->integer('sla_id')
            ->notEmptyString('sla_id');

        $validator
            ->numeric('minimal_availability_percent')
            ->requirePresence('minimal_availability_percent', 'create')
            ->notEmptyString('minimal_availability_percent');

        $validator
            ->numeric('determined_availability_percent')
            ->requirePresence('determined_availability_percent', 'create')
            ->notEmptyString('determined_availability_percent');

        $validator
            ->integer('total_time')
            ->requirePresence('total_time', 'create')
            ->notEmptyString('total_time');

        $validator
            ->integer('minimal_availability_time')
            ->requirePresence('minimal_availability_time', 'create')
            ->notEmptyString('minimal_availability_time');

        $validator
            ->integer('determined_availability_time')
            ->requirePresence('determined_availability_time', 'create')
            ->notEmptyString('determined_availability_time');

        $validator
            ->integer('determined_outage_time')
            ->requirePresence('determined_outage_time', 'create')
            ->notEmptyString('determined_outage_time');

        $validator
            ->integer('determined_number_outages')
            ->notEmptyString('determined_number_outages');

        $validator
            ->dateTime('evaluation_start')
            ->requirePresence('evaluation_start', 'create')
            ->notEmptyDateTime('evaluation_start');

        $validator
            ->dateTime('evaluation_end')
            ->requirePresence('evaluation_end', 'create')
            ->notEmptyDateTime('evaluation_end');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param \Cake\ORM\RulesChecker $rules The rules object to be modified.
     * @return \Cake\ORM\RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('host_id', 'Hosts'), ['errorField' => 'host_id']);
        $rules->add($rules->existsIn('service_id', 'Services'), ['errorField' => 'service_id']);
        $rules->add($rules->existsIn('sla_id', 'Slas'), ['errorField' => 'sla_id']);

        return $rules;
    }

    /**
     * @param $serviceId
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getSlaStatusLogDataByServiceId($serviceId, $slaId) {
        $query = $this->find()
            ->select([
                'SlaAvailabilityStatusServicesLog.id',
                'SlaAvailabilityStatusServicesLog.determined_availability_percent',
                'SlaAvailabilityStatusServicesLog.minimal_availability_percent',
                'SlaAvailabilityStatusServicesLog.evaluation_start',
            ])
            ->contain([
                'Services'          => function (Query $query) use ($serviceId, $slaId) {
                    return $query
                        ->select([
                            'Services.id'
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Services.id'       => $serviceId,
                            'Services.disabled' => 0
                        ]);
                },
                'SlaServiceOutages' => function (Query $query) use ($serviceId, $slaId) {
                    return $query
                        ->where([
                            'SlaServiceOutages.in_downtime' => 0
                        ]);
                }
            ])
            ->where([
                'SlaAvailabilityStatusServicesLog.service_id' => $serviceId,
                'SlaAvailabilityStatusServicesLog.sla_id'     => $slaId
            ])
            ->order(['SlaAvailabilityStatusServicesLog.id' => 'desc'])
            ->limit(10)
            ->disableHydration()
            ->all();

        if (empty($query)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param SlaStatusLogConditions $SlaStatusLogConditions
     * @param $PaginateOMat
     * @param $enableHydration
     * @return array
     */
    public function getSlaServiceStatusLogIndex(SlaStatusLogConditions $SlaStatusLogConditions, $PaginateOMat = null, $enableHydration = true) {
        $query = $this->find()
            ->contain(['SlaServiceOutages'])
            ->where([
                'SlaAvailabilityStatusServicesLog.service_id'         => $SlaStatusLogConditions->getId(),
                'SlaAvailabilityStatusServicesLog.evaluation_start >' => date('Y-m-d H:i:s', $SlaStatusLogConditions->getFrom()),
                'SlaAvailabilityStatusServicesLog.evaluation_start <' => date('Y-m-d H:i:s', $SlaStatusLogConditions->getTo())
            ])
            ->order($SlaStatusLogConditions->getOrder());

        if ($SlaStatusLogConditions->hasConditions()) {
            $query->andWhere($SlaStatusLogConditions->getConditions());
        }


        $query->enableHydration($enableHydration);

        if ($PaginateOMat === null) {
            //Just execute query
            $result = $this->emptyArrayIfNull($query->toArray());
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param $serviceId
     * @param $from
     * @param $to
     * @return array
     */
    public function getSlaServiceStatusLogByServiceIdWithRange($serviceId, $from, $to): array {
        $query = $this->find();
        $query->contain([
            'SlaServiceOutages'
        ]);

        $query->where([
            'SlaAvailabilityStatusServicesLog.service_id'          => $serviceId,
            'SlaAvailabilityStatusServicesLog.evaluation_start >=' => date('Y-m-d H:i:s', $from),
            'SlaAvailabilityStatusServicesLog.evaluation_end <='   => date('Y-m-d H:i:s', $to)
        ])
            ->order([
                'SlaAvailabilityStatusServicesLog.evaluation_start' => 'asc',
                'SlaAvailabilityStatusServicesLog.id'
            ])
            ->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }

    /**
     * @param $slaId
     * @param $from
     * @param $to
     * @return array
     */
    public function getLowestSlaServiceStatusLogBySlaIdWithRangeGroupByEvaluationEnd($slaId, $from, $to, $MY_RIGHTS = []): array {
        $query = $this->find('list', [
            'keyField'   => function ($row) {
                return $row['evaluation_start']->timestamp;
            },
            'valueField' => 'determined_availability_service'
        ]);
        $query->select([
            'SlaAvailabilityStatusServicesLog.evaluation_start',
            'determined_availability_service' => $query->func()
                ->min('SlaAvailabilityStatusServicesLog.determined_availability_percent')
        ]);

        if (!empty($MY_RIGHTS)) {
            $query
                ->contain('Services')
                ->innerJoinWith('Hosts')
                ->innerJoinWith('Hosts.HostsToContainersSharing', function (Query $q) use ($MY_RIGHTS) {
                    $q->where([
                        'HostsToContainersSharing.id IN ' => $MY_RIGHTS
                    ]);

                    return $q;
                });
        }
        $query->where([
            'SlaAvailabilityStatusServicesLog.sla_id'              => $slaId,
            'SlaAvailabilityStatusServicesLog.evaluation_start >=' => date('Y-m-d H:i:s', $from),
            'SlaAvailabilityStatusServicesLog.evaluation_start <='   => date('Y-m-d H:i:s', $to)
        ])
            ->order([
                'SlaAvailabilityStatusServicesLog.evaluation_start' => 'asc',
                'SlaAvailabilityStatusServicesLog.id'
            ])
            ->group('SlaAvailabilityStatusServicesLog.evaluation_end')
            ->disableHydration();

        return $this->emptyArrayIfNull($query->toArray());
    }
}
