<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.


declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PluginManagerTableTrait;
use App\Model\Table\HostsTable;
use App\Model\Table\ServicesTable;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\RepositoryInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\Validation\Validator;
use SLAModule\Model\Entity\SlaAvailabilityStatusService;

/**
 * SlaAvailabilityStatusServices Model
 *
 * @property HostsTable&BelongsTo $Hosts
 * @property ServicesTable&BelongsTo $Services
 * @property SlasTable&BelongsTo $Slas
 *
 * @method SlaAvailabilityStatusService newEmptyEntity()
 * @method SlaAvailabilityStatusService newEntity(array $data, array $options = [])
 * @method SlaAvailabilityStatusService[] newEntities(array $data, array $options = [])
 * @method SlaAvailabilityStatusService get($primaryKey, $options = [])
 * @method SlaAvailabilityStatusService findOrCreate($search, ?callable $callback = null, $options = [])
 * @method SlaAvailabilityStatusService patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method SlaAvailabilityStatusService[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method SlaAvailabilityStatusService|false save(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusService saveOrFail(EntityInterface $entity, $options = [])
 * @method SlaAvailabilityStatusService[]|\Cake\Datasource\ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusService[]|\Cake\Datasource\ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusService[]|\Cake\Datasource\ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method SlaAvailabilityStatusService[]|\Cake\Datasource\ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class SlaAvailabilityStatusServicesTable extends Table {

    use PluginManagerTableTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('sla_availability_status_services');
        $this->setDisplayField('id');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Hosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'Hosts',
        ]);
        $this->belongsTo('Services', [
            'foreignKey' => 'service_id',
            'joinType'   => 'INNER',
            'className'  => 'Services',
        ]);
        $this->belongsTo('Slas', [
            'foreignKey' => 'sla_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.Slas',
        ]);
        $this->belongsTo('SlaAvailabilityStatusHosts', [
            'foreignKey' => 'host_id',
            'joinType'   => 'INNER',
            'className'  => 'SLAModule.SlaAvailabilityStatusHosts',
        ]);

    }

    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Services':
                $coreTable->hasOne('SlaAvailabilityStatusServices', [
                    'className'        => 'SLAModule.SlaAvailabilityStatusServices',
                    'dependent'        => true,
                    'foreignKey'       => 'service_id',
                    'cascadeCallbacks' => true
                ]);
                break;
        }
    }


    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->integer('host_id')
            ->notEmptyString('host_id');

        $validator
            ->integer('service_id')
            ->notEmptyString('service_id');

        $validator
            ->integer('sla_id')
            ->notEmptyString('sla_id');

        $validator
            ->numeric('minimal_availability_percent')
            ->requirePresence('minimal_availability_percent', 'create')
            ->notEmptyString('minimal_availability_percent');

        $validator
            ->numeric('determined_availability_percent')
            ->requirePresence('determined_availability_percent', 'create')
            ->notEmptyString('determined_availability_percent');

        $validator
            ->integer('total_time')
            ->requirePresence('total_time', 'create')
            ->notEmptyString('total_time');

        $validator
            ->integer('minimal_availability_time')
            ->requirePresence('minimal_availability_time', 'create')
            ->notEmptyString('minimal_availability_time');

        $validator
            ->integer('determined_availability_time')
            ->requirePresence('determined_availability_time', 'create')
            ->notEmptyString('determined_availability_time');

        $validator
            ->integer('determined_outage_time')
            ->requirePresence('determined_outage_time', 'create')
            ->notEmptyString('determined_outage_time');

        $validator
            ->integer('determined_number_outages')
            ->notEmptyString('determined_number_outages');

        $validator
            ->dateTime('evaluation_start')
            ->requirePresence('evaluation_start', 'create')
            ->notEmptyDateTime('evaluation_start');

        $validator
            ->dateTime('evaluation_end')
            ->requirePresence('evaluation_end', 'create')
            ->notEmptyDateTime('evaluation_end');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('service_id', 'Services'), ['errorField' => 'service_id']);
        $rules->add($rules->existsIn('host_id', 'Hosts'), ['errorField' => 'host_id']);
        $rules->add($rules->existsIn('sla_id', 'Slas'), ['errorField' => 'sla_id']);

        return $rules;
    }

    public function getSlaStatusDataByHostIdAndSlaId($hostId, $slaId) {
        $result = $this->find()
            ->contain([
                'Services' => function (Query $query) use ($hostId, $slaId) {
                    $query = $query->contain([
                        'Servicetemplates' => function (Query $query) {
                            return $query->select([
                                'Servicetemplates.id'
                            ])->disableAutoFields();
                        }
                    ]);
                    return $query
                        ->select([
                            'Services.id',
                            'servicename' => $query->newExpr('IF(Services.name IS NULL, Servicetemplates.name, Services.name)')
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Services.disabled' => 0
                        ]);
                }
            ])->where([
                'SlaAvailabilityStatusServices.sla_id'  => $slaId,
                'SlaAvailabilityStatusServices.host_id' => $hostId
            ])->order([
                'SlaAvailabilityStatusServices.determined_availability_percent' => 'asc'
            ])->disableHydration()
            ->all();

        if (empty($result)) {
            return [];
        }
        return $result->toArray();
    }

    /**
     * @param $serviceId
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getSlaStatusDataByServiceId($serviceId, $slaId) {
        $query = $this->find()
            ->contain([
                'Services' => function (Query $query) use ($serviceId, $slaId) {
                    return $query
                        ->select([
                            'Services.id',
                            'Services.name'
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Services.id'       => $serviceId,
                            'Services.disabled' => 0
                        ]);
                }
            ])
            ->where([
                'SlaAvailabilityStatusServices.sla_id' => $slaId
            ])
            ->disableHydration();
        return $query->first();
    }

    /**
     * @param $hostId
     * @param $slaId
     * @param $evaluationStart
     * @return mixed
     */
    public function getServiceWithMinimalDeterminedAvailabilityByHostIdAndSlaId($hostId, $slaId, $evaluationStart) {
        $query = $this->find();
        $query->select([
            'min_determined_services_availability_percent' => $query->func()
                ->min('SlaAvailabilityStatusServices.determined_availability_percent')
        ])
            ->contain([
                'Services' => function (Query $query) use ($hostId) {
                    return $query
                        ->select([
                            'Services.id',
                            'Services.name'
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Services.host_id'  => $hostId,
                            'Services.disabled' => 0
                        ]);
                }
            ])
            ->where([
                'SlaAvailabilityStatusServices.host_id'          => $hostId,
                'SlaAvailabilityStatusServices.sla_id'           => $slaId,
                'SlaAvailabilityStatusServices.evaluation_start' => $evaluationStart
            ])
            ->disableHydration();
        return $query->first();
    }
}
