<?php
// Copyright (C) <2015-present>  <it-novum GmbH>
//
// This file is licensed under the terms of the openITCOCKPIT Enterprise Edition license agreement.
// The license agreement and license key were sent with the order confirmation.

declare(strict_types=1);

namespace SLAModule\Model\Table;

use App\Lib\Traits\PaginationAndScrollIndexTrait;
use App\Lib\Traits\PluginManagerTableTrait;
use App\Model\Table\ContainersTable;
use App\Model\Table\HostsTable;
use App\Model\Table\TimeperiodsTable;
use Cake\Database\Expression\Comparison;
use Cake\Datasource\EntityInterface;
use Cake\Datasource\RepositoryInterface;
use Cake\Datasource\ResultSetInterface;
use Cake\ORM\Association\BelongsTo;
use Cake\ORM\Behavior\TimestampBehavior;
use Cake\ORM\Query;
use Cake\ORM\RulesChecker;
use Cake\ORM\Table;
use Cake\ORM\TableRegistry;
use Cake\Validation\Validator;
use itnovum\openITCOCKPIT\Database\PaginateOMat;
use itnovum\openITCOCKPIT\Filter\GenericFilter;
use SLAModule\Model\Entity\Sla;

/**
 * Slas Model
 *
 * @property ContainersTable&BelongsTo $Containers
 * @property TimeperiodsTable&BelongsTo $Timeperiods
 *
 * @method Sla newEmptyEntity()
 * @method Sla newEntity(array $data, array $options = [])
 * @method Sla[] newEntities(array $data, array $options = [])
 * @method Sla get($primaryKey, $options = [])
 * @method Sla findOrCreate($search, ?callable $callback = null, $options = [])
 * @method Sla patchEntity(EntityInterface $entity, array $data, array $options = [])
 * @method Sla[] patchEntities(iterable $entities, array $data, array $options = [])
 * @method Sla|false save(EntityInterface $entity, $options = [])
 * @method Sla saveOrFail(EntityInterface $entity, $options = [])
 * @method Sla[]|ResultSetInterface|false saveMany(iterable $entities, $options = [])
 * @method Sla[]|ResultSetInterface saveManyOrFail(iterable $entities, $options = [])
 * @method Sla[]|ResultSetInterface|false deleteMany(iterable $entities, $options = [])
 * @method Sla[]|ResultSetInterface deleteManyOrFail(iterable $entities, $options = [])
 *
 * @mixin TimestampBehavior
 */
class SlasTable extends Table {
    use PaginationAndScrollIndexTrait;
    use PluginManagerTableTrait;

    /**
     * Initialize method
     *
     * @param array $config The configuration for the Table.
     * @return void
     */
    public function initialize(array $config): void {
        parent::initialize($config);

        $this->setTable('slas');
        $this->setDisplayField('name');
        $this->setPrimaryKey('id');

        $this->addBehavior('Timestamp');

        $this->belongsTo('Containers', [
            'foreignKey' => 'container_id',
            'joinType'   => 'INNER',
            'className'  => 'Containers',
        ]);
        $this->belongsTo('Timeperiods', [
            'foreignKey' => 'timeperiod_id',
            'joinType'   => 'INNER',
            'className'  => 'Timeperiods',
        ]);

        $this->hasOne('HostMappingRules', [
            'className' => 'SLAModule.HostMappingRules',
        ])->setDependent(true);


        $this->hasMany('Hosttemplates', [
            'foreignKey' => 'sla_id'
        ]);


        $this->hasMany('Hosts', [
            'foreignKey' => 'sla_id'
        ]);

        $this->hasMany('SlaAvailabilityStatusHosts', [
            'className'  => 'SLAModule.SlaAvailabilityStatusHosts',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);

        $this->hasMany('SlaAvailabilityStatusServices', [
            'className'  => 'SLAModule.SlaAvailabilityStatusServices',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);

        // Relations for clean deletion of SLAs
        $this->hasMany('SlaAvailabilityStatusHostsLog', [
            'className'  => 'SLAModule.SlaAvailabilityStatusHostsLog',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);

        $this->hasMany('SlaAvailabilityStatusServicesLog', [
            'className'  => 'SLAModule.SlaAvailabilityStatusServicesLog',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);

        $this->hasMany('SlaHostOutages', [
            'className'  => 'SLAModule.SlaHostOutages',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);

        $this->hasMany('SlaServiceOutages', [
            'className'  => 'SLAModule.SlaServiceOutages',
            'foreignKey' => 'sla_id',
            'dependent'  => true
        ]);
    }


    /**
     * Bind Core2Plugin associations
     *
     * @param RepositoryInterface $coreTable
     */
    public function bindCoreAssociations(RepositoryInterface $coreTable) {
        switch ($coreTable->getAlias()) {
            case 'Hosttemplates':
            case 'Hosts':
                $coreTable->belongsTo('Slas', [
                    'className' => 'SLAModule.Slas',
                    'dependent' => false
                ]);
                break;
        }
    }


    /**
     * Default validation rules.
     *
     * @param Validator $validator Validator instance.
     * @return Validator
     */
    public function validationDefault(Validator $validator): Validator {
        $validator
            ->scalar('name')
            ->maxLength('name', 255)
            ->requirePresence('name', 'create')
            ->notEmptyString('name');

        $validator
            ->scalar('description')
            ->maxLength('description', 255)
            ->allowEmptyString('description');

        $validator
            ->integer('container_id')
            ->notEmptyString('container_id');

        $validator
            ->integer('timeperiod_id')
            ->notEmptyString('timeperiod_id');

        $validator
            ->numeric('minimal_availability')
            ->requirePresence('minimal_availability', 'create')
            ->range('minimal_availability', [0, 100], 'Minimal availability must be in the range 0-100')
            ->notEmptyString('minimal_availability')
            ->lessThanField('minimal_availability', 'warning_threshold', 'Must be lesser than "Warning threshold"',
                function ($context) {
                    return !empty($context['data']['warning_threshold']);
                });

        $validator
            ->numeric('warning_threshold')
            ->range('warning_threshold', [0, 100], 'Warning threshold must be in the range 0-100')
            ->allowEmptyString('warning_threshold')
            ->greaterThanField('warning_threshold', 'minimal_availability', 'Must be greater than "Minimal availability"',
                function ($context) {
                    return !empty($context['data']['minimal_availability']);
                });

        $validator
            ->scalar('evaluation_interval')
            ->notEmptyString('evaluation_interval');

        $validator
            ->scalar('start_date')
            ->date('start_date', ['ymd'])
            ->allowEmptyString('start_date');

        return $validator;
    }

    /**
     * Returns a rules checker object that will be used for validating
     * application integrity.
     *
     * @param RulesChecker $rules The rules object to be modified.
     * @return RulesChecker
     */
    public function buildRules(RulesChecker $rules): RulesChecker {
        $rules->add($rules->existsIn('container_id', 'Containers'), ['errorField' => 'container_id']);
        $rules->add($rules->existsIn('timeperiod_id', 'Timeperiods'), ['errorField' => 'timeperiod_id']);

        return $rules;
    }

    /**
     * @param GenericFilter $GenericFilter
     * @param PaginateOMat|null $PaginateOMat
     * @param array $MY_RIGHTS
     * @return array
     */
    public function getSlasIndex(GenericFilter $GenericFilter, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        $query = $this->find('all')
            ->contain([
                'Timeperiods' => function (Query $query) {
                    $query
                        ->disableAutoFields()
                        ->select([
                            'id',
                            'name'
                        ]);
                    return $query;
                }
            ])
            ->disableResultsCasting()
            ->disableHydration();


        $indexFilter = [];
        if (!empty($MY_RIGHTS)) {
            $indexFilter['Slas.container_id IN'] = $MY_RIGHTS;
        }
        $query->where($indexFilter);

        if (!empty($GenericFilter->genericFilters())) {
            $query->where($GenericFilter->genericFilters());
        }

        $query->order($GenericFilter->getOrderForPaginator('Slas.name', 'asc'));
        if ($PaginateOMat === null) {
            //Just execute query
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }

        return $result;
    }

    /**
     * @param int $id
     * @return bool
     */
    public function existsById($id) {
        return $this->exists(['Slas.id' => $id]);
    }

    /**
     * @param array $containerIds
     * @param string $type
     * @param string $index
     * @return array
     */
    public function getSlasByContainerId($containerIds = [], $type = 'all', $index = 'id') {
        if (!is_array($containerIds)) {
            $containerIds = [$containerIds];
        }

        //Lookup for the tenant container of $container_id
        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');


        $tenantContainerIds = [];

        foreach ($containerIds as $containerId) {
            if ($containerId != ROOT_CONTAINER) {

                // Get container id of the tenant container
                // $container_id is may be a location, devicegroup or whatever, so we need to container id of the tenant container to load contactgroups and contacts
                $path = $ContainersTable->getPathByIdAndCacheResult($containerId, 'SlasByContainerId');

                // Tenant contact groups are available for all users of a tenant (oITC V2 legacy)
                if (isset($path[1])) {
                    $tenantContainerIds[] = $path[1]['id'];
                }
            } else {
                $tenantContainerIds[] = ROOT_CONTAINER;
            }
        }
        $tenantContainerIds = array_unique($tenantContainerIds);
        $containerIds = array_unique(array_merge($tenantContainerIds, $containerIds));


        if (empty($containerIds)) {
            return [];
        }

        $query = $this->find()
            ->where([
                'container_id IN' => $containerIds
            ])
            ->disableHydration()
            ->all();

        $records = $query->toArray();
        if (empty($records) || is_null($records)) {
            return [];
        }

        if ($type === 'all') {
            return $records;
        }

        $list = [];
        foreach ($records as $record) {
            $name = $record['name'];
            $list[$record[$index]] = $name;
        }
        return $list;
    }

    public function getHostsForOverviewBySlaId($id, HostsTable $HostsTable, GenericFilter $GenericFilter, PaginateOMat $PaginateOMat = null, $MY_RIGHTS = []) {
        //Lookup for the tenant container of $container_id
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');
        $query = $HostsTable->find();
        $query->select([
            'Hosts.id',
            'Hosts.name',
            'Hosts.address',
            'Hosts.sla_id',
            'Hosts.container_id',
            'container' => $query->newExpr('SELECT Containers.name')
        ]);
        if (!empty($MY_RIGHTS)) {
            $query->select([
                'primary_container' => $query->newExpr(
                    'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                AND containers_temp.id IN (' . implode(',', $MY_RIGHTS) . ')
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
            ]);
        } else {
            $query->select([
                'primary_container' => $query->newExpr(
                    'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')
            ]);
        }
        $query->innerJoinWith('Containers');

        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }

        $query->contain([
            'HostsToContainersSharing',
            'Hosttemplates' => [
                'fields' => [
                    'Hosttemplates.id',
                    'Hosttemplates.name',
                    'Hosttemplates.sla_id',
                    'Hosttemplates.container_id'
                ]
            ]
        ]);


        $where = $GenericFilter->genericFilters();
        $where[] = new Comparison(
            'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id)',
            $id,
            'integer',
            '='
        );

        if (isset($where['Hosts.name rlike'])) {
            if ($this->isRegularExpression($where['Hosts.name rlike'])) {
                $where[] = new Comparison(
                    'Hosts.name',
                    $where['Hosts.name rlike'],
                    'string',
                    'RLIKE'
                );
            }
            unset($where['Hosts.name rlike']);
        }

        $query->where($where);

        $query->disableHydration();
        $query->group(['Hosts.id']);

        $query->order(
            array_merge(
                $GenericFilter->getOrderForPaginator('Hosts.name', 'desc'),
                ['Hosts.id' => 'asc']
            )
        );

        if ($PaginateOMat === null) {
            //Just execute query
            if (empty($query)) {
                return [];
            }
            $result = $query->toArray();
        } else {
            if ($PaginateOMat->useScroll()) {
                $result = $this->scrollCake4($query, $PaginateOMat->getHandler());
            } else {
                $result = $this->paginateCake4($query, $PaginateOMat->getHandler());
            }
        }
        return $result;
    }

    /**
     * @param $string
     * @return bool
     */
    public function isRegularExpression($string): bool {
        return @preg_match('`' . $string . '`', '') !== false;
    }

    /**
     * @return Sla[]|ResultSetInterface
     */
    public function getSlasForCronjob() {
        $query = $this->find()
            ->contain([
                'Timeperiods'   => [
                    'TimeperiodTimeranges',
                    'Calendars' => 'CalendarHolidays'
                ],
                'Hosts',
                'Hosttemplates' => function (Query $q) {
                    return $q->enableAutoFields(false)
                        ->select([
                            'id',
                            'sla_id'
                        ])
                        ->contain([
                            'Hosts' => function (Query $query) {
                                $query
                                    ->disableAutoFields()
                                    ->select([
                                        'Hosts.id',
                                        'Hosts.uuid',
                                        'Hosts.name',
                                        'Hosts.hosttemplate_id'
                                    ]);
                                $query
                                    ->leftJoinWith('Slas')
                                    ->whereNull('Slas.id');
                                return $query;
                            }
                        ]);
                }
            ])
            ->order(['Slas.name' => 'asc']);
        $query->all();

        if (empty($query)) {
            return [];
        }
        return $query;
    }

    /**
     * @param int $slaId
     * @return ResultSetInterface
     */
    public function getHostsBySlaId(int $slaId): ResultSetInterface {
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $query = $HostsTable->find()
            ->select([
                'Hosts.id',
                'Hosts.uuid',
                'Hosts.name',
                'Hosts.created',
                'SlaAvailabilityStatusHosts.id',
                'SlaAvailabilityStatusHosts.host_id',
                'SlaAvailabilityStatusHosts.evaluation_end',
                'SlaAvailabilityStatusHosts.evaluation_total_time',
                'SlaAvailabilityStatusHosts.total_time',
                'SlaAvailabilityStatusHosts.determined_outage_time',
                'SlaAvailabilityStatusHosts.determined_availability_percent',
                'SlaAvailabilityStatusHosts.determined_number_outages'
            ])
            ->contain([
                'Hosttemplates',
                'SlaAvailabilityStatusHosts',
                'Services' => function (Query $query) {
                    return $query
                        ->select([
                            'Services.id',
                            'Services.uuid',
                            'Services.host_id',
                            'Services.created',
                            'SlaAvailabilityStatusServices.id',
                            'SlaAvailabilityStatusServices.evaluation_end',
                            'SlaAvailabilityStatusServices.evaluation_total_time',
                            'SlaAvailabilityStatusServices.total_time',
                            'SlaAvailabilityStatusServices.determined_outage_time',
                            'SlaAvailabilityStatusServices.determined_availability_percent',
                            'SlaAvailabilityStatusServices.determined_number_outages'
                        ])
                        ->contain([
                            'SlaAvailabilityStatusServices'
                        ])
                        ->where([
                            'Services.disabled' => 0
                        ]);
                }
            ])
            ->where(['Hosts.disabled' => 0])
            ->where(new Comparison(
                'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id)',
                $slaId,
                'integer',
                '='
            ));
        return $query->all();
    }

    /**
     * @param $hostId
     * @param $slaId
     * @return Query
     */
    public function getMinSlaStatusInformationByHostIdAndSlaId($hostId, $slaId) {
        $query = $this->find()
            ->contain([
                'SlaAvailabilityStatusHosts' => function (Query $query) use ($hostId, $slaId) {
                    return $query
                        ->select([
                            'Hosts.id',
                            'Hosts.sla_id',
                            'SlaAvailabilityStatusHosts.sla_id',
                            'SlaAvailabilityStatusHosts.determined_availability_percent',
                            'SlaAvailabilityStatusHosts.evaluation_start',
                            'SlaAvailabilityStatusHosts.evaluation_end'
                        ])
                        ->contain([
                            'Hosts' => [
                                'Services' => function (Query $query) {
                                    return $query
                                        ->select([
                                            'Services.host_id',
                                            'min_determined_services_availability_percent' => $query->func()->min(
                                                'SlaAvailabilityStatusServices.determined_availability_percent'
                                            ),
                                            'SlaAvailabilityStatusServices.evaluation_start',
                                            'SlaAvailabilityStatusServices.evaluation_end',
                                        ])
                                        ->contain([
                                            'SlaAvailabilityStatusServices'
                                        ])
                                        ->where([
                                            'Services.disabled' => 0
                                        ])
                                        ->disableAutoFields();
                                }
                            ]
                        ])
                        ->where([
                            'SlaAvailabilityStatusHosts.host_id' => $hostId,
                            'SlaAvailabilityStatusHosts.sla_id'  => $slaId,
                        ])
                        ->disableAutoFields();
                }
            ])
            ->where([
                'Slas.id' => $slaId
            ]);
        $query->disableHydration();
        return $query->first();
    }

    /**
     * @param $serviceId
     * @param $slaId
     * @return Query
     */
    public function getSlaStatusInformationByServiceIdAndSlaId($serviceId, $slaId) {
        $query = $this->find()
            ->contain([
                'SlaAvailabilityStatusServices' => function (Query $query) use ($serviceId) {
                    return $query
                        ->select([
                            'Services.id',
                            'SlaAvailabilityStatusServices.sla_id',
                            'SlaAvailabilityStatusServices.determined_availability_percent',
                            'SlaAvailabilityStatusServices.evaluation_start',
                            'SlaAvailabilityStatusServices.evaluation_end'

                        ])
                        ->contain([
                            'Services'
                        ])
                        ->disableAutoFields()
                        ->where([
                            'Services.id'       => $serviceId,
                            'Services.disabled' => 0
                        ]);
                }
            ])
            ->where([
                'Slas.id' => $slaId
            ])
            ->disableAutoFields()
            ->disableHydration();
        return $query->first();
    }

    /**
     * @param $slaId
     * @return array|EntityInterface|null
     */
    public function getSlaById($slaId) {
        return $this->find()
            ->contain([
                'Timeperiods' => 'TimeperiodTimeranges'
            ])
            ->where([
                'Slas.id' => $slaId
            ])
            ->disableResultsCasting()
            ->first();
    }

    public function deleteSlaRecordsByHostId($hostId) {
        return $this->deleteSlaRecordsByHostIds([
            (int)$hostId
        ]);
    }

    /**
     * Deletes host related records from ALL SLA tables.
     * This is used if a host got moved from one SLA to another.
     *
     * @param array $hostIds
     * @return bool
     */
    public function deleteSlaRecordsByHostIds(array $hostIds): bool {
        $tableNames = [
            'SLAModule.SlaAvailabilityStatusHosts',
            'SLAModule.SlaAvailabilityStatusServices',
            'SLAModule.SlaAvailabilityStatusHostsLog',
            'SLAModule.SlaAvailabilityStatusServicesLog',
            'SLAModule.SlaHostOutages',
            'SLAModule.SlaServiceOutages'
        ];

        $saveHostIds = [];
        foreach ($hostIds as $hostId) {
            $id = (int)$hostId;
            if ($id > 0) {
                $saveHostIds[] = $id;
            }
        }


        if (!empty($saveHostIds)) {
            foreach ($tableNames as $tableName) {
                /** @var Table $Table */
                $Table = TableRegistry::getTableLocator()->get($tableName);
                $Table->deleteAll([
                    'host_id IN' => $saveHostIds
                ]);
            }
        }

        return true;
    }


    /**
     * @param int $slaId
     * @param $MY_RIGHTS
     * @return ResultSetInterface
     */
    public function getHostsWithSlaStatusBySlaId(int $slaId, $MY_RIGHTS = []): ResultSetInterface {
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $query = $HostsTable->find();
        $query->select([
            'Hosts.id',
            'Hosts.name',
            'determined_availability' => $query->func()
                ->min(
                    $query->newExpr(
                        'IF(SlaAvailabilityStatusHosts.determined_availability_percent > 
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent,  
                    SlaAvailabilityStatusHosts.lowest_determined_availability_of_services_percent, 
                    SlaAvailabilityStatusHosts.determined_availability_percent)'
                    )
                )
        ])
            ->contain([
                'Hosttemplates',
                'SlaAvailabilityStatusHosts'
            ])
            ->where(['Hosts.disabled' => 0])
            ->where(new Comparison(
                'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id)',
                $slaId,
                'integer',
                '='
            ));
        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }
        $query->order([
            'determined_availability' => 'asc'
        ]);
        $query->group(['Hosts.id']);

        return $query->all();
    }

    /**
     * @param int $slaId
     * @param array $MY_RIGHTS
     * @param bool $considerContainer
     * @return ResultSetInterface
     */
    public function getHostsAndServicesWithSlaStatusBySlaId(int $slaId, array $MY_RIGHTS = [], bool $considerContainer = false): ResultSetInterface {
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $query = $HostsTable->find();
        $query->select([
            'Hosts.id',
            'Hosts.name',
            'Hosts.container_id',
            'SlaAvailabilityStatusHosts.determined_availability_percent'
        ])
            ->contain([
                'Hosttemplates',
                'SlaAvailabilityStatusHosts',
                'Services' => function (Query $query) {
                    return $query
                        ->select([
                            'Services.id',
                            'Services.host_id',
                            'SlaAvailabilityStatusServices.determined_availability_percent'
                        ])
                        ->contain([
                            'SlaAvailabilityStatusServices'
                        ])
                        ->where([
                            'Services.disabled' => 0
                        ]);
                }
            ])
            ->where(['Hosts.disabled' => 0])
            ->where(new Comparison(
                'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id)',
                $slaId,
                'integer',
                '='
            ));
        if ($considerContainer === true) {
            if (!empty($MY_RIGHTS)) {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                AND containers_temp.id IN (' . implode(',', $MY_RIGHTS) . ')
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')

                ]);
            } else {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                    (
                                        SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                        FROM containers AS containers_orig
                                        LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                        WHERE containers_orig.id = Containers.id
                                        ORDER BY containers_orig.lft ASC
                                    ), "/",
                                    Containers.name)
                                    )
                                ')

                ]);
            }
            $query->select([
                'Containers.id',
                'Containers.name'
            ])->contain('Containers');
        }
        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }

        return $query->all();
    }

    /**
     * @param int $slaId
     * @param $MY_RIGHTS
     * @param bool $considerContainerInformation
     * @return array
     */
    public
    function getAllHostsBySlaId(int $slaId, $MY_RIGHTS = [], bool $considerContainerInformation = false): array {
        /** @var HostsTable $HostsTable */
        $HostsTable = TableRegistry::getTableLocator()->get('Hosts');

        $query = $HostsTable->find();
        $query->select([
            'Hosts.id',
            'Hosts.name',
            'Hosts.created'
        ])
            ->contain([
                'Hosttemplates'
            ])
            ->where(['Hosts.disabled' => 0])
            ->where(new Comparison(
                'IF((Hosts.sla_id IS NULL), Hosttemplates.sla_id, Hosts.sla_id)',
                $slaId,
                'integer',
                '='
            ));
        if (!empty($MY_RIGHTS)) {
            $query->innerJoin(['HostsToContainersSharing' => 'hosts_to_containers'], [
                'HostsToContainersSharing.host_id = Hosts.id'
            ]);
            $query->where([
                'HostsToContainersSharing.container_id IN' => $MY_RIGHTS
            ]);
        }
        if ($considerContainerInformation === true) {
            $query->innerJoinWith('Containers');
            if (!empty($MY_RIGHTS)) {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                            (
                                                SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                                FROM containers AS containers_orig
                                                LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                                WHERE containers_orig.id = Containers.id
                                                AND containers_temp.id IN (' . implode(',', $MY_RIGHTS) . ')
                                                ORDER BY containers_orig.lft ASC
                                            ), "/",
                                            Containers.name)
                                            )
                                        ')

                ]);
            } else {
                $query->select([
                    'primary_container' => $query->newExpr(
                        'SELECT IF(Containers.name = "root", "root", CONCAT(
                                    (
                                        SELECT GROUP_CONCAT(containers_temp.name ORDER BY containers_temp.lft ASC SEPARATOR "/")
                                        FROM containers AS containers_orig
                                        LEFT JOIN containers AS containers_temp ON containers_temp.lft < containers_orig.lft AND containers_temp.rght > containers_orig.rght
                                        WHERE containers_orig.id = Containers.id
                                        ORDER BY containers_orig.lft ASC
                                    ), "/",
                                    Containers.name)
                                    )
                                ')

                ]);
            }
        }
        $query->order([
            'Hosts.name',
            'Hosts.id'
        ]);
        $query->disableHydration();
        if (empty($query)) {
            return [];
        }
        return $query->toArray();
    }

    /**
     * @param $containerIds
     * @return array
     */
    public
    function getSlasByContainerIds($containerIds = []) {
        if (!is_array($containerIds)) {
            $containerIds = [$containerIds];
        }

        //Lookup for the tenant container of $container_id
        /** @var $ContainersTable ContainersTable */
        $ContainersTable = TableRegistry::getTableLocator()->get('Containers');


        $tenantContainerIds = [];

        foreach ($containerIds as $containerId) {
            if ($containerId != ROOT_CONTAINER) {

                // Get container id of the tenant container
                // $container_id is may be a location, devicegroup or whatever, so we need to container id of the tenant container to load contactgroups and contacts
                $path = $ContainersTable->getPathByIdAndCacheResult($containerId, 'SlasByContainerId');

                // Tenant contact groups are available for all users of a tenant (oITC V2 legacy)
                if (isset($path[1])) {
                    $tenantContainerIds[] = $path[1]['id'];
                }
            } else {
                $tenantContainerIds[] = ROOT_CONTAINER;
            }
        }
        $tenantContainerIds = array_unique($tenantContainerIds);
        $containerIds = array_unique(array_merge($tenantContainerIds, $containerIds));


        if (empty($containerIds)) {
            return [];
        }

        $query = $this->find()
            ->where([
                'container_id IN' => $containerIds
            ])
            ->disableResultsCasting()
            ->all();

        if ($query === null) {
            return [];
        }
        return $query;
    }
}
